""":mod:`libearth.compat.parallel` --- Threading-related compatibility layer
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. function:: cpu_count()

   Get the number of CPU cores.

   :returns: the number of cpu cores
   :rtype: :class:`numbers.Integral`

.. function:: parallel_map(pool_size, function, iterable, *iterables)

   Parallel vesion of builtin :func:`map()` except of some differences:

   - It takes a more argument at first: ``pool_size``.
   - The function applications will be done in parallel.
   - The order of arguments to results are not maintained.
     You should treat these as a set.
   - The result is a lazy iterable.  Although the function immediately returns
     an iterable, it might block if some results are not completely ready
     when it's iterated.

   :param pool_size: the number of workers
   :type pool_size: :class:`numbers.Integral`
   :param function: the function to apply iterables as its arguments
   :type function: :class:`collections.Callable`
   :param iterable: function argument values
   :type iterable: :class:`collections.Iterable`
   :returns: a promise iterable to future results
   :rtype: :class:`collections.Iterable`

   .. versionchanged:: 0.1.1
      Errored values are raised at the lastest.

"""
import collections
import numbers
import sys

from . import IRON_PYTHON, PY3

__all__ = 'cpu_count', 'parallel_map'


if IRON_PYTHON:
    from System import Environment

    def cpu_count():
        return Environment.ProcessorCount
else:
    import multiprocessing

    def cpu_count():
        try:
            return multiprocessing.cpu_count()
        except NotImplementedError:
            return 1


try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    if IRON_PYTHON:
        from System import Action
        from System.Collections.Concurrent import BlockingCollection
        from System.Collections.Generic import IEnumerable
        from System.Threading import Thread, ThreadStart
        from System.Threading.Tasks import Parallel, ParallelOptions

        class parallel_map(collections.Iterable):

            ForEach = Parallel.ForEach[object].Overloads[
                IEnumerable[object],
                ParallelOptions,
                Action[object]
            ]

            def __init__(self, pool_size, function, *iterables):
                if not isinstance(pool_size, numbers.Integral):
                    raise TypeError('pool_size must be an integer, not ' +
                                    repr(pool_size))
                elif not callable(function):
                    raise TypeError('function must be callable, not ' +
                                    repr(function))
                elif not iterables:
                    raise TypeError('missing iterable')
                self.function = function
                self.results = BlockingCollection[tuple]()
                args = zip(*iterables)
                self.length = len(args)
                options = ParallelOptions()
                options.MaxDegreeOfParallelism = pool_size
                Thread.__new__.Overloads[ThreadStart](
                    Thread,
                    lambda: self.ForEach(args, options, self.store_result)
                ).Start()

            def store_result(self, args):
                try:
                    value = self.function(*args)
                except Exception as e:
                    result = None, e
                else:
                    result = value, None
                self.results.Add(result)

            def __iter__(self):
                errors = []
                for _ in xrange(self.length):
                    value, error = self.results.Take()
                    if error is None:
                        yield value
                    else:
                        errors.append(error)
                for error in errors:
                    raise error
    else:
        from multiprocessing.pool import ThreadPool

        class parallel_map(collections.Iterable):

            def __init__(self, pool_size, function, *iterables):
                if not isinstance(pool_size, numbers.Integral):
                    raise TypeError('pool_size must be an integer, not ' +
                                    repr(pool_size))
                elif not callable(function):
                    raise TypeError('function must be callable, not ' +
                                    repr(function))
                elif not iterables:
                    raise TypeError('missing iterable')
                self.pool = ThreadPool(pool_size)
                self.function = function
                self.results = self.pool.imap_unordered(self.map_function,
                                                        zip(*iterables))

            def map_function(self, args):
                try:
                    value = self.function(*args)
                except Exception:
                    return False, sys.exc_info()
                return True, value

            def __iter__(self):
                errors = []
                for success, value in self.results:
                    if success:
                        yield value
                    else:
                        errors.append(value)
                self.pool.close()
                self.pool.join()
                for error in errors:
                    exec('raise error[1], None, error[2]')
else:
    class parallel_map(collections.Iterable):

        def __init__(self, pool_size, function, *iterables):
            if not isinstance(pool_size, numbers.Integral):
                raise TypeError('pool_size must be an integer, not ' +
                                repr(pool_size))
            elif not callable(function):
                raise TypeError('function must be callable, not ' +
                                repr(function))
            elif not iterables:
                raise TypeError('missing iterable')
            self.pool = ThreadPoolExecutor(pool_size)
            self.function = function
            self.results = self.pool.map(self.map_function, *iterables)

        def map_function(self, *args):
            try:
                value = self.function(*args)
            except Exception:
                return False, sys.exc_info()
            return True, value

        def __iter__(self):
            errors = []
            for success, value in self.results:
                if success:
                    yield value
                else:
                    errors.append(value)
            self.pool.shutdown()
            if PY3:
                for _, exc, tb in errors:
                    raise exc.with_traceback(tb)
            else:
                for _, exc, tb in errors:
                    exec('raise exc, None, tb')



""":mod:`libearth.compat.etree` --- ElementTree compatibility layer
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This proxy module offers a compatibility layer between several ElementTree
implementations.

- If there's installed :mod:`lxml` module, use :mod:`lxml.etree`.
- If :mod:`xml.etree.cElementTree` is available, use it.
- If IronPython, use :mod:`xml.etree.ElementTree` with
  :class:`libearth.compat.clrxmlreader.TreeBuilder`.
- Otherwise, use :mod:`xml.etree.ElementTree`.

It provides the following two functions:

.. function:: fromstring(string)

   Parse the given XML ``string``.

   :param string: xml string to parse
   :type string: :class:`str`, :class:`bytes`, :class:`basestring`
   :returns: the element tree object

.. function:: fromstringlist(iterable)

   Parse the given chunks of XML string.

   :param iterable: chunks of xml string to parse
   :type iterable: :class:`collections.Iterable`
   :returns: the element tree object

.. function:: tostring(tree)

   Generate an XML string from the given element tree.

   :param tree: an element tree object to serialize
   :returns: an xml string
   :rtype: :class:`str`, :class:`bytes`

"""
try:
    from lxml.etree import fromstring, fromstringlist, tostring
except ImportError:
    try:
        from xml.etree.cElementTree import fromstring, tostring
    except ImportError:
        from xml.etree.ElementTree import fromstring, tostring
        try:
            from xml.etree.ElementTree import fromstringlist
        except ImportError:
            fromstringlist = None
    else:
        try:
            from xml.etree.cElementTree import fromstringlist
        except ImportError:
            fromstringlist = None

from . import IRON_PYTHON

__all__ = 'fromstring', 'fromstringlist'


if IRON_PYTHON:
    from xml.etree import ElementTree
    from .clrxmlreader import TreeBuilder

    def fromstring(string, parser=None):
        if parser is None:
            parser = TreeBuilder()
        return ElementTree.fromstring(string, parser)

    def fromstringlist(iterable, parser=None):
        if parser is None:
            parser = TreeBuilder()
        return ElementTree.fromstringlist(iterable, parser)
elif fromstringlist is None:
    def fromstringlist(iterable):
        it = iter(iterable)
        try:
            first = next(it)
        except StopIteration:
            return fromstring('')
        return fromstring(first + type(first)().join(it))



""":mod:`libearth.repository` --- Repositories
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:class:`Repository` abstracts storage backend e.g. filesystem.
There might be platforms that have no chance to directly access
file system e.g. iOS, and in that case the concept of repository
makes you to store data directly to Dropbox_ or `Google Drive`_
instead of filesystem.  However in the most cases we will simply use
:class:`FileSystemRepository` even if data are synchronized using
Dropbox or :program:`rsync`.

In order to make the repository highly configurable it provides the way
to lookup and instantiate the repository from url.  For example,
the following url will load :class:`FileSystemRepository` which sets
:attr:`~FileSystemRepository.path` to :file:`/home/dahlia/.earthreader/`:

.. code-block:: text

   file:///home/dahlia/.earthreader/

For extensibility every repository class has to implement :meth:`from_url()`
and :meth:`to_url()` methods, and register it as an `entry point`__ of
``libearth.repositories`` group e.g.:

.. code-block:: ini

   [libearth.repositories]
   file = libearth.repository:FileSystemRepository

Note that the entry point name (``file`` in the above example) becomes
the url scheme to lookup the corresponding repository class
(:class:`libearth.repository.FileSystemRepository` in the above example).

.. _Dropbox: http://dropbox.com/
.. _Google Drive: https://drive.google.com/
__ https://pythonhosted.org/setuptools/pkg_resources.html#entry-points

"""
import collections
import errno
import io
import os
import os.path
import pipes
import shutil
import sys
import tempfile
import threading
try:
    from urllib import parse as urlparse
except ImportError:
    import urlparse
import weakref

from .compat import IRON_PYTHON, string_type, xrange

__all__ = ('FileIterator', 'FileNotFoundError', 'FileSystemRepository',
           'NotADirectoryError', 'Repository', 'RepositoryKeyError',
           'from_url')


def from_url(url):
    """Load the repository instance from the given configuration ``url``.

    .. note::

       If :mod:`setuptools` is not installed it will only support
       ``file://`` scheme and :class:`FileSystemRepository`.

    :param url: a repository configuration url
    :type url: :class:`str`, :class:`urllib.parse.ParseResult`
    :returns: the loaded repository instance
    :rtype: :class:`Repository`
    :raises LookupError: when the corresponding repository type to
                         the given ``url`` scheme cannot be found
    :raises ValueError: when the given ``url`` is invalid

    """
    if isinstance(url, string_type):
        url = urlparse.urlparse(url)
    elif not isinstance(url, urlparse.ParseResult):
        raise TypeError(
            'url must be a string, or an instance of {0.__module__}.'
            '{0.__name__}, not {1!r}'.format(urlparse.ParseResult, url)
        )
    lookup_error = LookupError('cannot find the corresponding repository to '
                               '{0}:// scheme'.format(url.scheme))
    try:
        from pkg_resources import iter_entry_points
    except ImportError:
        if url.scheme != 'file':  # FIXME
            raise lookup_error
        repository_type = FileSystemRepository
    else:
        entry_points = iter_entry_points(
            group='libearth.repositories',
            name=url.scheme
        )
        for ep in entry_points:
            repository_type = ep.load()
            if issubclass(repository_type, Repository):
                break
        else:
            raise lookup_error
    return repository_type.from_url(url)


class Repository(object):
    """Repository interface agnostic to its underlying storage implementation.
    :class:`~libearth.stage.Stage` objects can deal with documents to be stored
    using the interface.

    Every content in repositories is accessible using *keys*.  It actually
    abstracts out "filenames" in "file systems", hence keys share the common
    concepts with filenames.  Keys are hierarchical, like file paths, so
    consists of multiple sequential strings e.g. ``['dir', 'subdir', 'key']``.
    You can :meth:`list()` all subkeys in the upper key as well e.g.::

        repository.list(['dir', 'subdir'])

    """

    @classmethod
    def from_url(cls, url):
        """Create a new instance of the repository from the given ``url``.
        It's used for configuring the repository in plain text
        e.g. :file:`*.ini`.

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`from_url()` static/class method to implement details.

        :param url: the parsed url tuple
        :type url: :class:`urllib.parse.ParseResult`
        :returns: a new repository instance
        :rtype: :class:`Repository`
        :raises ValueError: when the given url is not invalid

        """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement from_url() static/class method'.format(Repository)
        )

    def to_url(self, scheme):
        """Generate a url that :meth:`from_url()` can accept.
        It's used for configuring the repository in plain text
        e.g. :file:`*.ini`.  URL ``scheme`` is determined by caller,
        and given through argument.

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`to_url()` method to implement details.

        :param scheme: a determined url scheme
        :returns: a url that :meth:`from_url()` can accept
        :rtype: :class:`str`

        """
        if not isinstance(scheme, string_type):
            raise TypeError('scheme must be a string, not ' + repr(scheme))
        if hash(type(self).to_url) == hash(Repository.to_url):
            raise NotImplementedError(
                'every subclass of {0.__module__}.{0.__name__} has to '
                'implement to_url() method'.format(Repository)
            )

    def read(self, key):
        """Read the content from the ``key``.

        :param key: the key which stores the content to read
        :type key: :class:`collections.Sequence`
        :returns: byte string chunks
        :rtype: :class:`collections.Iterable`
        :raises RepositoryKeyError: the ``key`` cannot be found in
                                    the repository, or it's not a file

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`read()` method to implement details.

        """
        if not isinstance(key, collections.Sequence):
            raise TypeError('key must be a sequence, not ' + repr(key))
        elif not key:
            raise RepositoryKeyError(key, 'key cannot be empty')
        if hash(type(self).read) == hash(Repository.read):
            raise NotImplementedError(
                'every subclass of {0.__module__}.{0.__name__} has to '
                'implement read() method'.format(Repository)
            )

    def write(self, key, iterable):
        """Write the ``iterable`` into the ``key``.

        :param key: the key to stores the ``iterable``
        :type key: :class:`collections.Sequence`
        :param iterable: the iterable object yiels chunks of the whole
                         content.  every chunk has to be a byte string
        :type iterable: :class:`collections.Iterable`

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`write()` method to implement details.

        """
        if not isinstance(key, collections.Sequence):
            raise TypeError('key must be a sequence, not ' + repr(key))
        elif not isinstance(iterable, collections.Iterable):
            raise TypeError('expected an iterable object, not ' +
                            repr(iterable))
        elif not key:
            raise RepositoryKeyError(key, 'key cannot be empty')
        if hash(type(self).write) == hash(Repository.write):
            raise NotImplementedError(
                'every subclass of {0.__module__}.{0.__name__} has to '
                'implement write() method'.format(Repository)
            )

    def exists(self, key):
        """Return whether the ``key`` exists or not.  It returns :const:`False`
        if it doesn't exist instead of raising :exc:`RepositoryKeyError`.

        :param key: the key to find whether it exists
        :type key: :class:`collections.Sequence`
        :returns: :const:`True` only if the given ``key`` exists,
                  or :const:`False` if not exists
        :rtype: :class:`bool`

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`exists()` method to implement details.

        """
        if not isinstance(key, collections.Sequence):
            raise TypeError('key must be a sequence, not ' + repr(key))
        if hash(type(self).exists) == hash(Repository.exists):
            raise NotImplementedError(
                'every subclass of {0.__module__}.{0.__name__} has to '
                'implement exists() method'.format(Repository)
            )

    def list(self, key):
        """List all subkeys in the ``key``.

        :param key: the incomplete key that might have subkeys
        :type key: :class:`collections.Sequence`
        :returns: the set of subkeys (set of strings, not set of string lists)
        :rtype: :class:`collections.Set`
        :raises RepositoryKeyError: the ``key`` cannot be found in
                                    the repository, or it's not a directory

        .. note::

           Every subclass of :class:`Repository` has to override
           :meth:`list()` method to implement details.

        """
        if not isinstance(key, collections.Sequence):
            raise TypeError('key must be a sequence, not ' + repr(key))
        if hash(type(self).list) == hash(Repository.list):
            raise NotImplementedError(
                'every subclass of {0.__module__}.{0.__name__} has to '
                'implement list() method'.format(Repository)
            )

    def __repr__(self):
        return '{0.__module__}.{0.__name__}()'.format(type(self))


class RepositoryKeyError(LookupError, IOError):
    """Exception which rises when the requested key cannot be found
    in the repository.

    """

    #: (:class:`collections.Sequence`) The requested key.
    key = None

    def __init__(self, key, *args, **kwargs):
        super(RepositoryKeyError, self).__init__(*args, **kwargs)
        self.key = key


class FileSystemRepository(Repository):
    """Builtin implementation of :class:`Repository` interface which uses
    the ordinary file system.

    :param path: the directory path to store keys
    :type path: :class:`str`
    :param mkdir: create the directory if it doesn't exist yet.
                  :const:`True` by default
    :type mkdir: :class:`bool`
    :param atomic: make the update invisible until it's complete.
                   :const:`False` by default
    :raises FileNotFoundError: when the ``path`` doesn't exist
    :raises NotADirectoryError: when the ``path`` is not a directory

    """

    #: (:class:`str`) The path of the directory to read and write data files.
    #: It should be readable and writable.
    path = None

    @classmethod
    def from_url(cls, url):
        if not isinstance(url, urlparse.ParseResult):
            raise TypeError(
                'url must be an instance of {0.__module__}.{0.__name__}, '
                'not {1!r}'.format(urlparse.ParseResult, url)
            )
        if url.scheme != 'file':
            raise ValueError('{0.__module__}.{0.__name__} only accepts '
                             'file:// scheme'.format(FileSystemRepository))
        elif url.netloc or url.params or url.query or url.fragment:
            raise ValueError('file:// must not contain any host/port/user/'
                             'password/parameters/query/fragment')
        if sys.platform == 'win32':
            if not url.path.startswith('/'):
                raise ValueError('invalid file path: ' + repr(url.path))
            parts = url.path.lstrip('/').split('/')
            path = os.path.join(parts[0] + os.path.sep, *parts[1:])
        else:
            path = url.path
        return cls(path)

    def __init__(self, path, mkdir=True, atomic=IRON_PYTHON):
        if not os.path.exists(path):
            if mkdir:
                try:
                    os.makedirs(path)
                except OSError as e:
                    if e.errno == errno.EEXIST:
                        pass
                    else:
                        raise
            else:
                raise FileNotFoundError(repr(path) + ' does not exist')
        if not os.path.isdir(path):
            raise NotADirectoryError(repr(path) + ' is not a directory')
        self.path = path
        self.atomic = atomic
        self.lock = threading.RLock()
        self.file_iterators = {}

    def to_url(self, scheme):
        super(FileSystemRepository, self).to_url(scheme)
        if sys.platform == 'win32':
            drive, path = os.path.splitdrive(self.path)
            path = '/'.join(path.lstrip(os.path.sep).split(os.path.sep))
            return '{0}:///{1}/{2}'.format(scheme, drive, path)
        return '{0}://{1}'.format(scheme, self.path)

    def read(self, key):
        super(FileSystemRepository, self).read(key)
        path = os.path.join(self.path, *key)
        if not os.path.isfile(path):
            raise RepositoryKeyError(key)
        with self.lock:
            iterator = FileIterator(path, buffer_size=4096)
            try:
                iterator_set = self.file_iterators[path]
            except KeyError:
                # weakref.WeakSet was introduced since Python 2.7,
                # so workaround it on Python 2.6 by using WeakKeyDictionary
                iterator_set = weakref.WeakKeyDictionary({iterator: 1})
                self.file_iterators[path] = iterator_set
            else:
                iterator_set[iterator] = len(iterator_set) + 1
            return iterator

    def write(self, key, iterable):
        super(FileSystemRepository, self).write(key, iterable)
        dirpath = list(key)[:-1]
        dirpath.insert(0, self.path)
        for i in xrange(len(dirpath)):
            p = os.path.join(*dirpath[:i + 1])
            if not os.path.exists(p):
                try:
                    os.mkdir(p)
                except OSError as e:
                    if e.errno == errno.EEXIST:
                        pass
                    else:
                        raise
            elif not os.path.isdir(p):
                raise RepositoryKeyError(key)
        filename = os.path.join(self.path, *key)
        with self.lock:
            already_opened_iterators = self.file_iterators.get(filename, {})
            for iterator in already_opened_iterators.keys():
                iterator.preload_all()
        if self.atomic:
            f = tempfile.NamedTemporaryFile('wb', delete=False)
        else:
            f = io.open(filename, 'wb')
        with f:
            for chunk in iterable:
                f.write(chunk)
        if self.atomic:
            if IRON_PYTHON:
                # FIXME: no mv in windows
                cmd = '/bin/mv {0} {1}'.format(
                    pipes.quote(f.name),
                    pipes.quote(filename)
                )
                with os.popen(cmd) as pf:
                    pf.read()
            else:
                shutil.move(f.name, filename)

    def exists(self, key):
        super(FileSystemRepository, self).exists(key)
        return os.path.exists(os.path.join(self.path, *key))

    def list(self, key):
        super(FileSystemRepository, self).list(key)
        try:
            names = os.listdir(os.path.join(self.path, *key))
        except (IOError, OSError) as e:
            raise RepositoryKeyError(key, str(e))
        return frozenset(name for name in names if name != '..' or name != '.')

    def __repr__(self):
        return '{0.__module__}.{0.__name__}({1!r})'.format(type(self),
                                                           self.path)


class FileIterator(collections.Iterator):
    """Read a file through :class:`~collections.Iterator` protocol,
    with automatic closing of the file when it ends.

    :param path: the path of file
    :type path: :class:`str`
    :param buffer_size: the size of bytes that would be produced each step
    :type buffer_size: :class:`numbers.Integral`

    """

    def __init__(self, path, buffer_size):
        self.path = path
        self.buffer_size = buffer_size
        self.file_ = None

    def __iter__(self):
        self.file_ = io.open(self.path, 'rb', buffering=0)
        return self

    def __next__(self):
        f = self.file_
        if f is None:
            f = self.__iter__().file_
        elif f.closed:
            if hasattr(self, 'preloaded'):
                rest = self.preloaded
                del self.preloaded
                return rest
            raise StopIteration
        try:
            chunk = f.read(self.buffer_size)
        except:
            self.file_.close()
            raise
        if chunk:
            return chunk
        self.file_.close()
        raise StopIteration

    next = __next__

    def tell(self):
        return self.file_ and self.file_.tell()

    def seek(self, *args):
        if self.file_ is not None:
            self.file_.seek(*args)

    def read(self, *args):
        if self.file_ is not None:
            return self.file_.read(*args)

    def preload_all(self):
        f = self.file_
        if f is None:
            f = self.__iter__().file_
        elif not f.closed:
            self.preloaded = f.read()
            f.close()


try:
    FileNotFoundError = FileNotFoundError
    NotADirectoryError = NotADirectoryError
except NameError:
    class FileNotFoundError(IOError, OSError):
        """Raised when a given path does not exist."""

    class NotADirectoryError(IOError, OSError):
        """Raised when a given path is not a directory."""



""":mod:`libearth.xml.compat.clrxmlreader` --- XML parser implementation for CLR
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Python :mod:`xml.sax` parser implementation and ElementTree builder using
CLR ``System.Xml.XmlReader``.

.. seealso::

   - `XmlReader Class`__
   - `Comparing XmlReader to SAX Reader`__

   __ http://msdn.microsoft.com/en-us/library/system.xml.xmlreader.aspx
   __ http://msdn.microsoft.com/en-us/library/sbw89de7.aspx

"""
import clr
clr.AddReference('System.Xml')
import collections
import System
import System.IO
import System.Text
import System.Xml
import xml.etree.ElementTree
import xml.sax.xmlreader

from .xmlpullreader import PullReader

__all__ = ('XMLNS_XMLNS', 'IteratorStream', 'TreeBuilder', 'XmlReader',
           'create_parser')


#: (:class:`str`) The reserved namespace URI for XML namespace.
XMLNS_XMLNS = 'http://www.w3.org/2000/xmlns/'


def create_parser():
    """Create a new :class:`XmlReader()` parser instance.

    :returns: a new parser instance
    :rtype: :class:`XmlReader`

    """
    return XmlReader()


class XmlReader(PullReader):
    """SAX :class:`~libearth.compat.xmlpullreader.PullReader` implementation
    using CLR ``System.Xml.XmlReader``.

    """

    def prepareParser(self, iterable):
        stream = IteratorStream(iterable)
        self.reader = System.Xml.XmlReader.Create(stream)

    def feed(self):
        reader = self.reader
        if not reader.Read():
            return False
        handler = self.getContentHandler()
        XmlNodeType = System.Xml.XmlNodeType
        node_type = reader.NodeType
        if node_type == XmlNodeType.Element:
            attrs = {}
            qnames = {}
            if reader.HasAttributes:
                while reader.MoveToNextAttribute():
                    if reader.NamespaceURI == XMLNS_XMLNS:
                        continue
                    attr = reader.NamespaceURI or None, reader.LocalName
                    attrs[attr] = reader.Value
                    qnames[attr] = reader.Name
                reader.MoveToElement()
            handler.startElementNS(
                (reader.NamespaceURI or None, reader.LocalName),
                reader.Name,
                xml.sax.xmlreader.AttributesNSImpl(attrs, qnames)
            )
            if reader.IsEmptyElement:
                handler.endElementNS(
                    (reader.NamespaceURI or None, reader.LocalName),
                    reader.Name
                )
        elif node_type == XmlNodeType.Text or node_type == XmlNodeType.CDATA:
            handler.characters(reader.Value)
        elif node_type == XmlNodeType.EndElement:
            handler.endElementNS(
                (reader.NamespaceURI or None, reader.LocalName),
                reader.Name
            )
        return True

    def close(self):
        self.reader.Dispose()

    def setFeature(self, name, state):
        pass

    def reset(self):
        pass


class IteratorStream(System.IO.Stream):
    """``System.IO.Stream`` implementation that takes a Python iterable
    and then transforms it into CLR stream.

    :param iterable: a Python iterable to transform
    :type iterable: :class:`collections.Iterable`

    """

    def __new__(cls, iterable):
        stream = System.IO.Stream.__new__(cls)
        stream.__init__(iterable)
        return stream

    def __init__(self, iterable):
        self.iterable = iterable
        self.iterator = iter(iterable)
        self.buffer = []
        self.buffer_length = 0
        self.position = 0
        self.length = 0
        self.length_cache = None

    def consume(self):
        try:
            chunk = next(self.iterator)
        except StopIteration:
            return False
        assert isinstance(chunk, bytes), repr(chunk) + ' is not bytes'
        chunk_size = len(chunk)
        self.buffer.append(chunk)
        self.buffer_length += chunk_size
        self.length += chunk_size
        return True

    def get_CanRead(self):
        return True

    def get_CanSeek(self):
        return False

    def get_CanWrite(self):
        return False

    def get_Length(self):
        if self.length_cache is not None:
            return self.length_cache
        if isinstance(self.iterable, collections.Sequence):
            self.length_cache = sum(len(bytes_) for bytes_ in self.iterable)
            return self.length_cache
        while self.consume():
            pass
        self.length_cache = self.length
        return self.length

    def get_Position(self):
        return self.position

    def Read(self, buffer, offset, count):
        while self.buffer_length < count:
            if not self.consume():
                break
        total_read = 0
        for i, chunk in enumerate(self.buffer):
            read_bytes = min(count, len(chunk))
            chunk[:read_bytes].CopyTo(buffer, offset + total_read)
            count -= read_bytes
            total_read += read_bytes
            if not count:
                break
        if total_read:
            if len(chunk) <= read_bytes:
                self.buffer[:i + 1] = []
            else:
                self.buffer[i] = chunk[read_bytes:]
                self.buffer[:i] = []
            self.buffer_length -= total_read
            self.position += total_read
        return total_read

    def Seek(self, offset, seek_origin):
        raise System.InvalidOperationException(
            'Seek is unsupported on this stream'
        )

    def SetLength(self, value):
        raise System.InvalidOperationException(
            'SetLength is unsupported on this stream'
        )

    def Write(self, buffer, offset, count):
        raise System.InvalidOperationException(
            'Write is unsupported on this stream'
        )


class TreeBuilder(object):
    """ElementTree builder using ``System.Xml.XmlReader``."""

    def __init__(self):
        self.buffer = System.IO.MemoryStream()
        self.encode = System.Text.Encoding.Default.GetBytes

    def feed(self, data):
        if isinstance(data, str):
            data = self.encode(data)
        else:
            data = data.ToByteArray()
        self.buffer.Write(data, 0, len(data))

    def close(self):
        self.buffer.Seek(0, System.IO.SeekOrigin.Begin)
        builder = xml.etree.ElementTree.TreeBuilder()
        reader = System.Xml.XmlReader.Create(self.buffer)
        XmlNodeType = System.Xml.XmlNodeType
        fixname = self.fixname
        while reader.Read():
            node_type = reader.NodeType
            if node_type == XmlNodeType.Element:
                attrib = {}
                if reader.HasAttributes:
                    while reader.MoveToNextAttribute():
                        if reader.NamespaceURI == XMLNS_XMLNS:
                            continue
                        attrib[fixname(reader)] = reader.Value
                    reader.MoveToElement()
                builder.start(fixname(reader), attrib)
                if reader.IsEmptyElement:
                    builder.end(fixname(reader))
            elif (node_type == XmlNodeType.Text or
                  node_type == XmlNodeType.CDATA):
                builder.data(reader.Value)
            elif node_type == XmlNodeType.EndElement:
                builder.end(fixname(reader))
        return builder.close()

    @staticmethod
    def fixname(reader):
        name = reader.LocalName
        xmlns = reader.NamespaceURI
        if xmlns:
            return '{' + xmlns + '}' + name
        return name



""":mod:`libearth.parser.autodiscovery` --- Autodiscovery
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides functions to autodiscovery feed url in document.

"""
try:
    import HTMLParser
except ImportError:
    import html.parser as HTMLParser
import collections
import logging
import re
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from ..compat import text
from ..compat.etree import fromstring
from .atom import parse_atom
from .rss2 import parse_rss
from .util import normalize_xml_encoding


__all__ = ('ATOM_TYPE', 'RSS_TYPE', 'TYPE_TABLE', 'AutoDiscovery', 'FeedLink',
           'FeedUrlNotFoundError', 'autodiscovery', 'get_format')


#: (:class:`str`) The MIME type of RSS 2.0 format
#: (:mimetype:`application/rss+xml`).
RSS_TYPE = 'application/rss+xml'

#: (:class:`str`) The MIME type of Atom format
#: (:mimetype:`application/atom+xml`).
ATOM_TYPE = 'application/atom+xml'

#: (:class:`collections.Set`) The set of supported feed MIME types.
#:
#: .. versionadded:: 0.3.0
FEED_TYPES = frozenset([RSS_TYPE, ATOM_TYPE])

#: (:class:`collections.Mapping`) The mapping table of feed types
TYPE_TABLE = {parse_atom: ATOM_TYPE, parse_rss: RSS_TYPE}

#: Namedtuple which is a pair of ``type` and ``url``
FeedLink = collections.namedtuple('FeedLink', 'type url')


def autodiscovery(document, url):
    """If the given url refers an actual feed, it returns the given url
    without any change.

    If the given url is a url of an ordinary web page
    (i.e. :mimetype:`text/html`), it finds the urls of the corresponding feed.
    It returns feed urls in feed types' lexicographical order.

    If autodiscovery failed, it raise :exc:`FeedUrlNotFoundError`.

    :param document: html, or xml strings
    :type document: :class:`str`
    :param url: the url used to retrieve the ``document``.
                if feed url is in html and represented in relative url,
                it will be rebuilt on top of the ``url``
    :type url: :class:`str`
    :returns: list of :class:`FeedLink` objects
    :rtype: :class:`collections.MutableSequence`

    """
    document = text(document)
    document_type = get_format(document)
    if document_type is None:
        parser = AutoDiscovery()
        feed_links, _ = parser.find(document)
        if not feed_links:
            raise FeedUrlNotFoundError('Cannot find feed url')
        for link in feed_links:
            if link.url.startswith('/'):
                absolute_url = urlparse.urljoin(url, link.url)
                feed_links[feed_links.index(link)] = \
                    FeedLink(link.type, absolute_url)
        return feed_links
    else:
        return [FeedLink(TYPE_TABLE[document_type], url)]


class AutoDiscovery(HTMLParser.HTMLParser):
    """Parse the given HTML and try finding the actual feed urls from it.

    .. versionchanged:: 0.3.0
       It became to find icon links as well, and :meth:`find_feed_url()`
       method (that returned only feed links) was gone, instead :meth:`find()`
       (that return a pair of feed links and icon links) was introduced.

    """

    LINK_PATTERN = re.compile(r'''rel\s?=\s?(?:'|")?([^'">]+)''')
    LINK_HREF_PATTERN = re.compile(r'''href\s?=\s?(?:'|")?([^'"\s>]+)''')
    LINK_TYPE_PATTERN = re.compile(r'''type\s?=\s?(?:'|")?([^'"\s>]+)''')

    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self.feed_links = []
        self.icon_links = []

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if not (tag == 'link' and 'rel' in attrs and 'href' in attrs):
            return
        if attrs['rel'] == 'alternate' and 'type' in attrs and \
           attrs['type'] in FEED_TYPES:
            self.feed_links.append(FeedLink(attrs['type'], attrs['href']))
        elif 'icon' in attrs['rel'].split():
            self.icon_links.append(attrs['href'])

    def find(self, document):
        document = text(document)
        match = re.match('.+</head>', document)
        if match:
            head = match.group(0)
        else:
            head = document
        chunks = re.findall('[^>]*(?:>|$)', head)
        for chunk in chunks:
            try:
                self.feed(chunk)
            except Exception:
                self.find_link_with_regex(chunk)
        self.feed_links = sorted(self.feed_links, key=lambda link: link.type)
        return self.feed_links, self.icon_links

    def find_link_with_regex(self, chunk):
        match = self.LINK_PATTERN.search(chunk)
        if not match:
            return
        href_match = self.LINK_HREF_PATTERN.search(chunk)
        if not href_match:
            return
        rels = match.group(1).split()
        href = href_match.group(1)
        if 'alternate' in rels:
            type_match = self.LINK_TYPE_PATTERN.search(chunk)
            if type_match:
                type_ = type_match.group(1)
                if type_ in FEED_TYPES:
                    self.feed_links.append(FeedLink(type_, href))
        if 'icon' in rels:
            self.icon_links.append(href)


class FeedUrlNotFoundError(Exception):
    """Exception raised when feed url cannot be found in html."""

    def __init__(self, msg):
        self.msg = msg


def get_format(document):
    """Guess the syndication format of an arbitrary ``document``.

    :param document: document string to guess
    :type document: :class:`str`, :class:`bytes`
    :returns: the function possible to parse the given ``document``
    :rtype: :class:`collections.Callable`

    .. versionchanged:: 0.2.0
       The function was in :mod:`libearth.parser.heuristic` module (which is
       removed now) before 0.2.0, but now it's moved to
       :mod:`libearth.parser.autodiscovery`.

    """
    document = normalize_xml_encoding(document)
    try:
        root = fromstring(document)
    except Exception as e:
        logger = logging.getLogger(__name__ + '.get_format')
        logger.debug('document = %r', document)
        logger.warning(e, exc_info=True)
        return None
    if root.tag in ('{http://www.w3.org/2005/Atom}feed',
                    '{http://purl.org/atom/ns#}feed'):
        return parse_atom
    elif root.tag == 'rss':
        return parse_rss
    else:
        return None



""":mod:`libearth.schema` --- Declarative schema for pulling DOM parser of XML
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

There are well-known two ways to parse XML:

Document Object Model
   It reads the whole XML and then makes a tree in memory.  You can easily
   treverse the document as a tree, but the parsing can't be streamed.
   Moreover it uses memory for data you don't use.

Simple API for XML
   It's an event-based sequential access parser.  It means you need to
   listen events from it and then utilize its still unstructured data
   by yourself.  In other words, you don't need to pay memory to data
   you never use if you simply do nothing for them when you listen
   the event.

Pros and cons between these two ways are obvious, but there could be
another way to parse XML: *mix them*.

The basic idea of this pulling DOM parser (which this module implements)
is that the parser can consume the stream just in time when you actually
reach the child node.  There should be an assumption for that: parsed XML
has a schema for it.  If the document is schema-free, this heuristic approach
loses the most of its efficiency.

So the parser should have the information about the schema of XML document
it'd parser, and we can declare the schema by defining classes.  It's a thing
like ORM for XML.  For example, suppose there is a small XML document:

.. code-block:: xml

   <?xml version="1.0"?>
   <person version="1.0">
     <name>Hong Minhee</name>
     <url>http://dahlia.kr/</url>
     <url>https://github.com/dahlia</url>
     <url>https://bitbucket.org/dahlia</url>
     <dob>1988-08-04</dob>
   </person>

You can declare the schema for this like the following class definition::

    class Person(DocumentElement):
        __tag__ = 'person'
        format_version = Attribute('version')
        name = Text('name')
        url = Child('url', URL, multiple=True)
        dob = Child('dob', Date)

"""
import collections
import copy
import inspect
import itertools
import operator
import platform
import weakref
import xml.sax
import xml.sax.handler
import xml.sax.saxutils

from .compat import UNICODE_BY_DEFAULT, binary_type, string_type
from .compat.xmlpullreader import PullReader

__all__ = ('PARSER_LIST', 'SCHEMA_XMLNS',
           'Attribute', 'Child', 'Codec', 'CodecDescriptor', 'CodecError',
           'Content', 'ContentHandler', 'DecodeError', 'Descriptor',
           'DescriptorConflictError', 'DocumentElement', 'Element',
           'ElementList', 'EncodeError', 'IntegrityError',
           'SchemaError', 'Text',
           'complete', 'element_list_for',
           'index_descriptors', 'inspect_attributes', 'inspect_child_tags',
           'inspect_content_tag', 'inspect_xmlns_set', 'is_partially_loaded',
           'read', 'validate', 'write')


#: (:class:`str`) The XML namespace name used for schema metadatq.
SCHEMA_XMLNS = 'http://earthreader.org/schema/'


class SchemaError(TypeError):
    """Error which rises when a schema definition has logical errors."""


class DescriptorConflictError(SchemaError, AttributeError):
    """Error which rises when a schema has duplicate descriptors more than
    one for the same attribute, the same child element, or the text node.

    """


class IntegrityError(SchemaError, AttributeError):
    """Rise when an element is invalid according to the schema."""


class CodecError(SchemaError, ValueError):
    """Rise when encoding/decoding between Python values and XML data
    goes wrong.

    """


class EncodeError(CodecError):
    """Rise when encoding Python values into XML data goes wrong."""


class DecodeError(CodecError):
    """Rise when decoding XML data to Python values goes wrong."""


class Descriptor(object):
    """Abstract base class for :class:`Child` and :class:`Text`."""

    #: (:class:`str`) The tag name.
    tag = None

    #: (:class:`str`) The optional XML namespace URI.
    xmlns = None

    #: (:class:`tuple`) The pair of (:attr:`xmlns`, :attr:`tag`).
    key_pair = None

    #: (:class:`bool`) Whether it is required for the element.
    #: If it's :const:`True` :attr:`multiple` has to be :const:`False`.
    required = None

    #: (:class:`bool`) Whether it can be zero or more for the element.
    #: If it's :const:`True` :attr:`required` has to be :const:`False`.
    multiple = None

    #: (:class:`collections.Callable`) An optional function to be used
    #: for sorting multiple elements.  It has to take an element and
    #: return a value for sort key.  It is the same to ``key`` option of
    #: :func:`sorted()` built-in function.
    #:
    #: It's available only when :attr:`multiple` is :const:`True`.
    #:
    #: Use :attr:`sort_reverse` for descending order.
    #:
    #: .. note::
    #:
    #:    It doesn't guarantee that all elements must be sorted in
    #:    runtime, but all elements become sorted when it's written
    #:    using :func:`write()` function.
    sort_key = None

    #: (:class:`bool`) Whether to reverse elements when they become
    #: sorted.  It is the same to ``reverse`` option of :func:`sorted()`
    #: built-in function.
    #:
    #: It's available only when :attr:`sort_key` is present.
    sort_reverse = None

    def __init__(self, tag, xmlns=None, required=False, multiple=False,
                 sort_key=None, sort_reverse=None):
        global _descriptor_counter
        if required and multiple:
            raise TypeError('required and multiple are exclusive')
        elif not multiple and sort_key is not None:
            raise TypeError('sort_key function can be used only for multiple '
                            'children')
        elif not (sort_key is None or callable(sort_key)):
            raise TypeError('sort_key function must be callable, not ' +
                            repr(sort_key))
        elif sort_key is None and sort_reverse is not None:
            raise TypeError('sort_reverse option is available only when '
                            'sort_key is also present')
        self.tag = tag
        self.xmlns = xmlns
        self.key_pair = self.xmlns, self.tag
        self.required = bool(required)
        self.multiple = bool(multiple)
        self.sort_key = sort_key
        self.sort_reverse = bool(sort_reverse)
        try:
            _descriptor_counter += 1
        except NameError:
            _descriptor_counter = 1
        self.descriptor_counter = _descriptor_counter

    def __get__(self, obj, cls=None):
        if isinstance(obj, Element):
            if self.multiple:
                return ElementList(obj, self)
            root = obj._root() if hasattr(obj, '_root') else None
            if root is not None and getattr(root, '_handler', None):
                handler = root._handler
                stack = handler.stack
                while ((obj._data.get(self) is None and
                       (not stack or stack[-1]))):
                    if not root._parse_next():
                        break
            return obj._data.get(self)
        return self

    def start_element(self, element, attribute):
        """Abstract method that is invoked when the parser meets a start
        of an element related to the descriptor.  It will be called by
        :class:`ContentHandler`.

        :param element: the parent element of the read element
        :type element: :class:`Element`
        :param attribute: the attribute name of the descriptor
        :type attribute: :class:`str`
        :returns: a value to reserve.  it will be passed to
                  ``reserved_value`` parameter of :meth:`end_element()`

        """
        raise NotImplementedError(
            'start_element() method has to be implemented'
        )

    def end_element(self, reserved_value, content):
        """Abstract method that is invoked when the parser meets an end
        of an element related to the descriptor.  It will be called by
        :class:`ContentHandler`.

        :param reserved_value: the value :meth:`start_element()` method
                               returned
        :param content: the content text of the read element
        :type content: :class:`str`

        """
        raise NotImplementedError('end_element() method has to be implemented')


class Child(Descriptor):
    """Declare a possible child element as a descriptor.

    In order to have :class:`Child` of the element type which is not
    defined yet (or self-referential) pass the class name of the element
    type to contain.  The name will be lazily evaluated e.g.::

        class Person(Element):
            '''Everyone can have their children, that also are a Person.'''

            children = Child('child', 'Person', multiple=True)

    :param tag: the tag name
    :type tag: :class:`str`
    :param xmlns: an optional XML namespace URI
    :type xmlns: :class:`str`
    :param element_type: the type of child element(s).
                         it has to be a subtype of :class:`Element`.
                         if it's a string it means referring the class name
                         which is going to be lazily evaluated
    :type element_type: :class:`type`, :class:`str`
    :param required: whether the child is required or not.
                     it's exclusive to ``multiple``.
                     :const:`False` by default
    :type required: :class:`bool`
    :param multiple: whether the child can be multiple.
                     it's exclusive to ``required``.
                     :const:`False` by default
    :type multiple: :class:`bool`
    :param sort_key: an optional function to be used for sorting
                     multiple child elements.  it has to take a child as
                     :class:`Element` and return a value for sort key.
                     it is the same to ``key`` option of :func:`sorted()`
                     built-in function.
                     note that *it doesn't guarantee that all elements must
                     be sorted in runtime*, but all elements become sorted
                     when it's written using :func:`write()` function.
                     it's available only when ``multiple`` is :const:`True`.
                     use ``sort_reverse`` for descending order.
    :type sort_key: :class:`collections.Callable`
    :param sort_reverse: ehether to reverse elements when they become
                         sorted.  it is the same to ``reverse`` option of
                         :func:`sorted()` built-in function.
                         it's available only when ``sort_key`` is present
    :type sort_reverse: :class:`bool`

    .. todo::

       It crashes when the document has any non-ASCII characters.

    """

    def __init__(self, tag, element_type, xmlns=None, required=False,
                 multiple=False, sort_key=None, sort_reverse=None):
        if isinstance(element_type, type):
            if not issubclass(element_type, Element):
                raise TypeError(
                    'element_type must be a subtype of {0.__module__}.'
                    '{0.__name__}, not {1!r}'.format(Element, element_type)
                )
        elif isinstance(element_type, string_type):
            frame = inspect.currentframe().f_back
            element_type = element_type, frame.f_locals, frame.f_globals
        else:
            raise TypeError('element_type must be a class (or a string '
                            'referring its name), not ' + repr(element_type))
        super(Child, self).__init__(
            tag,
            xmlns=xmlns,
            required=required,
            multiple=multiple,
            sort_key=sort_key,
            sort_reverse=sort_reverse
        )
        self._element_type = element_type

    @property
    def element_type(self):
        """(:class:`type`) The class of this child can contain.  It must
        be a subtype of :class:`Element`.

        """
        if not isinstance(self._element_type, type):
            name, loc, glob = self._element_type
            try:
                self._element_type = loc[name]
            except KeyError:
                try:
                    self._element_type = glob[name]
                except KeyError:
                    raise NameError('name {0!r} is not defined'.format(name))
        return self._element_type

    def __get__(self, obj, cls=None):
        if isinstance(obj, Element):
            if self.multiple:
                return ElementList(obj, self, self.element_type)
        return super(Child, self).__get__(obj, cls)

    def __set__(self, obj, value):
        if isinstance(obj, Element):
            element_type = self.element_type
            if self.multiple:
                if isinstance(value, collections.Sequence):
                    if len(value) < 1 or \
                       isinstance(value[0], element_type):
                        value = [e if isinstance(e, element_type)
                                 else element_type.__coerce_from__(e)
                                 for e in value]
                        for e in value:
                            if not isinstance(e, element_type):
                                raise TypeError(
                                    'expected instances of {0.__module__}.'
                                    '{0.__name__}, not {1!r}'.format(e)
                                )
                        obj._data[self] = value
                    else:
                        raise TypeError(
                            'expected a sequence of {0.__module__}.'
                            '{0.__name__}, not {1!r}'.format(
                                element_type, value
                            )
                        )
                else:
                    raise TypeError('Child property of multiple=True option '
                                    'only accepts a sequence, not ' +
                                    repr(value))
            else:
                if not (value is None or isinstance(value, element_type)):
                    value = element_type.__coerce_from__(value)
                if not (value is None or isinstance(value, element_type)):
                    raise TypeError(
                        'expected an instance of {0.__module__}.{0.__name__}, '
                        'not {1!r}'.format(element_type, value)
                    )
                obj._data[self] = value
        else:
            raise AttributeError('cannot change the class attribute')

    def start_element(self, element, attribute):
        child_element = self.element_type(element)
        if self.multiple:
            element_list = element._data.setdefault(self, [])
            element_list.append(child_element)
        else:
            if self not in element._data:
                setattr(element, attribute, child_element)
        return child_element

    def end_element(self, reserved_value, content):
        element_type = type(reserved_value)
        attr = inspect_content_tag(element_type)
        if attr is not None:
            content_desc = attr[1]
            content_desc.read(reserved_value, content)
        reserved_value._partial = 0


class Codec(object):
    """Abstract base class for codecs to serialize Python values to be
    stored in XML and deserialize XML texts to Python values.

    In most cases encoding and decoding are implementation details of
    *format* which is well-defined, so these two functions could be
    paired.  The interface rely on that idea.

    To implement a codec, you have to subclass :class:`Codec` and
    override a pair of methods: :meth:`encode()` and :meth:`decode()`.

    Codec objects are acceptable by :class:`Attribute`, :class:`Text`, and
    :class:`Content` (all they subclass :class:`CodecDescriptor`).

    """

    def encode(self, value):
        """Encode the given Python ``value`` into XML text.

        :param value: Python value to encode
        :returns: the encoded XML text
        :rtype: :class:`str`
        :raise EncodeError: when encoding the given ``value`` goes wrong

        .. note::

           Every :class:`Codec` subtype has to override this method.

        """
        cls = type(self)
        raise NotImplementedError('{0.__module__}.{0.__name__}.encode() '
                                  'is not implemented'.format(cls))

    def decode(self, text):
        """Decode the given XML ``text`` to Python value.

        :param text: XML text to decode
        :type text: :class:`str`
        :returns: the decoded Python value
        :raise DecodeError: when decoding the given XML ``text`` goes wrong

        .. note::

           Every :class:`Codec` subtype has to override this method.

        """
        cls = type(self)
        raise NotImplementedError('{0.__module__}.{0.__name__}.decode() '
                                  'is not implemented'.format(cls))


# Semi-structured record type for only internal use.
CodecFunction = collections.namedtuple('CodecFunction', 'function descriptor')


class CodecDescriptor(object):
    """Mixin class for descriptors that provide :meth:`decoder` and
    :meth:`encoder`.

    :class:`Attribute`, :class:`Content` and :class:`Text` can take
    ``encoder`` and ``decoder`` functions for them.  It's used for encoding
    from Python values to XML string and decoding raw values from XML to
    natural Python representations.

    It can take a ``codec``, or ``encode`` and ``decode`` separately.
    (Of course they all can be present at a time.)  In most cases,
    you'll need only ``codec`` parameter that encoder and decoder
    are coupled::

        Text('dob', Rfc3339(prefer_utc=True))

    Encoders can be specified using ``encoder`` parameter of descriptor's
    constructor, or :meth:`encoder()` decorator.

    Decoders can be specified using ``decoder`` parameter of descriptor's
    constructor, or :meth:`decoder()` decorator::

        class Person(DocumentElement):
            __tag__ = 'person'
            format_version = Attribute('version')
            name = Text('name')
            url = Child('url', URL, multiple=True)
            dob = Text('dob',
                       encoder=datetime.date.strftime.isoformat,
                       decoder=lambda s: datetime.date.strptime(s, '%Y-%m-%d'))

            @format_version.encoder
            def format_version(self, value):
                return '.'.join(map(str, value))

            @format_version.decoder
            def format_version(self, value):
                return tuple(map(int, value.split('.')))

    :param codec: an optional codec object to use.  if it's callable and
                  not an instance of :class:`Codec`, its return value will
                  be used instead.  it means this can take class object of
                  :class:`Codec` subtype that is not instantiated yet
                  unless the constructor require any arguments
    :type codec: :class:`Codec`, :class:`collections.Callable`
    :param encoder: an optional function that encodes Python value into
                    XML text value e.g. :func:`str()`.  the encoder function
                    has to take an argument
    :type encoder: :class:`collections.Callable`
    :param decoder: an optional function that decodes XML text value into
                    Python value e.g. :func:`int()`.  the decoder function
                    has to take a string argument
    :type decoder: :class:`collections.Callable`

    """

    def __init__(self, codec=None, encoder=None, decoder=None):
        self.encoders = []
        self.decoders = []
        if callable(codec) and not isinstance(codec, Codec):
            codec = codec()
        if codec is not None:
            if not isinstance(codec, Codec):
                raise TypeError('codec must be an instance of {0.__module__}.'
                                '{0.__name__}, not {1!r}'.format(Codec, codec))
            self.encoders.append(CodecFunction(codec.encode, descriptor=False))
            self.decoders.append(CodecFunction(codec.decode, descriptor=False))
        if encoder is not None:
            if not callable(encoder):
                raise TypeError('encoder must be callable, not ' +
                                repr(encoder))
            self.encoders.append(CodecFunction(encoder, descriptor=False))
        if decoder is not None:
            if not callable(decoder):
                raise TypeError('decoder must be callable, not ' +
                                repr(decoder))
            self.decoders.append(CodecFunction(decoder, descriptor=False))

    def encoder(self, function):
        r"""Decorator which sets the encoder to the decorated function::

            import datetime

            class Person(DocumentElement):
                '''Person.dob will be written to ISO 8601 format'''

                __tag__ = 'person'
                dob = Text('dob')

                @dob.encoder
                def dob(self, dob):
                    if not isinstance(dob, datetime.date):
                        raise TypeError('expected datetime.date')
                    return dob.strftime('%Y-%m-%d')

        >>> isinstance(p, Person)
        True
        >>> p.dob
        datetime.date(1987, 7, 26)
        >>> ''.join(write(p, indent='', newline=''))
        '<person><dob>1987-07-26</dob></person>'

        If it's applied multiple times, all decorated functions are piped
        in the order::

            class Person(Element):
                '''Person.email will have mailto: prefix when it's written
                to XML.

                '''

                email = Text('email', encoder=lambda email: 'mailto:' + email)

                @age.encoder
                def email(self, email):
                    return email.strip()

                @email.encoder
                def email(self, email):
                    login, host = email.split('@', 1)
                    return login + '@' + host.lower()

        >>> isinstance(p, Person)
        True
        >>> p.email
        '  earthreader@librelist.com  '
        >>> ''.join(write(p, indent='', newline=''))
        >>> '<person><email>mailto:earthreader@librelist.com</email></person>')

        .. note::

           This creates a copy of the descriptor instance rather than
           manipulate itself in-place.

        """
        desc = copy.copy(self)
        desc.encoders = self.encoders[:]
        desc.encoders.append(CodecFunction(function, descriptor=True))
        return desc

    def decoder(self, function):
        r"""Decorator which sets the decoder to the decorated function::

            import datetime

            class Person(DocumentElement):
                '''Person.dob will be a datetime.date instance.'''

                __tag__ = 'person'
                dob = Text('dob')

                @dob.decoder
                def dob(self, dob_text):
                    return datetime.date.strptime(dob_text, '%Y-%m-%d')

        >>> p = Person('<person><dob>1987-07-26</dob></person>')
        >>> p.dob
        datetime.date(1987, 7, 26)

        If it's applied multiple times, all decorated functions are piped
        in the order::

            class Person(Element):
                '''Person.age will be an integer.'''

                age = Text('dob', decoder=lambda text: text.strip())

                @age.decoder
                def age(self, dob_text):
                    return datetime.date.strptime(dob_text, '%Y-%m-%d')

                @age.decoder
                def age(self, dob):
                    now = datetime.date.today()
                    d = now.month < dob.month or (now.month == dob.month and
                                                  now.day < dob.day)
                    return now.year - dob.year - d

        >>> p = Person('<person>\n\t<dob>\n\t\t1987-07-26\n\t</dob>\n</person>')
        >>> p.age
        26
        >>> datetime.date.today()
        datetime.date(2013, 7, 30)

        .. note::

           This creates a copy of the descriptor instance rather than
           manipulate itself in-place.

        """
        desc = copy.copy(self)
        desc.decoders = self.decoders[:]
        desc.decoders.append(CodecFunction(function, descriptor=True))
        return desc

    def encode(self, value, instance):
        for encoder in self.encoders:
            if encoder.descriptor and hasattr(encoder.function, '__get__'):
                value = encoder.function.__get__(instance)(value)
            else:
                value = encoder.function(value)
        return value

    def decode(self, text, instance):
        """Decode the given ``text`` as it's programmed.

        :param text: the raw text to decode.  xml attribute value or
                     text node value in most cases
        :type text: :class:`str`
        :param instance: the instance that is associated with the descriptor
        :type instance: :class:`Element`
        :returns: decoded value

        .. note::

           Internal method.

        """
        for decoder in self.decoders:
            if decoder.descriptor and hasattr(decoder.function, '__get__'):
                text = decoder.function.__get__(instance)(text)
            else:
                text = decoder.function(text)
        return text


class Text(Descriptor, CodecDescriptor):
    """Descriptor that declares a possible child element that only cosists
    of character data.  All other attributes and child nodes are ignored.

    :param tag: the XML tag name
    :type tag: :class:`str`
    :param codec: an optional codec object to use.  if it's callable and
                  not an instance of :class:`Codec`, its return value will
                  be used instead.  it means this can take class object of
                  :class:`Codec` subtype that is not instantiated yet
                  unless the constructor require any arguments
    :type codec: :class:`Codec`, :class:`collections.Callable`
    :param xmlns: an optional XML namespace URI
    :type xmlns: :class:`str`
    :param required: whether the child is required or not.
                     it's exclusive to ``multiple``.
                     :const:`False` by default
    :type required: :class:`bool`
    :param multiple: whether the child can be multiple.
                     it's exclusive to ``required``.
                     :const:`False` by default
    :type multiple: :class:`bool`
    :param encoder: an optional function that encodes Python value into
                    XML text value e.g. :func:`str()`.  the encoder function
                    has to take an argument
    :type encoder: :class:`collections.Callable`
    :param decoder: an optional function that decodes XML text value into
                    Python value e.g. :func:`int()`.  the decoder function
                    has to take a string argument
    :type decoder: :class:`collections.Callable`
    :param sort_key: an optional function to be used for sorting
                     multiple child elements.  it has to take a child as
                     :class:`Element` and return a value for sort key.
                     it is the same to ``key`` option of :func:`sorted()`
                     built-in function.
                     note that *it doesn't guarantee that all elements must
                     be sorted in runtime*, but all elements become sorted
                     when it's written using :func:`write()` function.
                     it's available only when ``multiple`` is :const:`True`.
                     use ``sort_reverse`` for descending order.
    :type sort_key: :class:`collections.Callable`
    :param sort_reverse: ehether to reverse elements when they become
                         sorted.  it is the same to ``reverse`` option of
                         :func:`sorted()` built-in function.
                         it's available only when ``sort_key`` is present
    :type sort_reverse: :class:`bool`

    """

    def __init__(self, tag, codec=None, xmlns=None, required=False,
                 multiple=False, encoder=None, decoder=None,
                 sort_key=None, sort_reverse=None):
        Descriptor.__init__(self, tag,
                            xmlns=xmlns, required=required, multiple=multiple,
                            sort_key=sort_key, sort_reverse=sort_reverse)
        CodecDescriptor.__init__(self, codec=codec,
                                 encoder=encoder, decoder=decoder)

    def __get__(self, obj, cls=None):
        if isinstance(obj, Element):
            if self.multiple:
                return ElementList(obj, self, string_type)
        return super(Text, self).__get__(obj, cls)

    def start_element(self, element, attribute):
        return element

    def end_element(self, reserved_value, content):
        content = self.decode(content, reserved_value)
        if self.multiple:
            reserved_value._data.setdefault(self, []).append(content)
        else:
            reserved_value._data.setdefault(self, content)


class Attribute(CodecDescriptor):
    """Declare possible element attributes as a descriptor.

    :param name: the XML attribute name
    :type name: :class:`str`
    :param codec: an optional codec object to use.  if it's callable and
                  not an instance of :class:`Codec`, its return value will
                  be used instead.  it means this can take class object of
                  :class:`Codec` subtype that is not instantiated yet
                  unless the constructor require any arguments
    :type codec: :class:`Codec`, :class:`collections.Callable`
    :param xmlns: an optional XML namespace URI
    :type xmlns: :class:`str`
    :param required: whether the child is required or not.
                     :const:`False` by default
    :type required: :class:`bool`
    :param default: an optional function that returns default value when
                    the attribute is not present.  the function takes an
                    argument which is an :class:`Element` instance
    :type default: :class:`collections.Callable`
    :param encoder: an optional function that encodes Python value into
                    XML text value e.g. :func:`str()`.  the encoder function
                    has to take an argument
    :type encoder: :class:`collections.Callable`
    :param decoder: an optional function that decodes XML text value into
                    Python value e.g. :func:`int()`.  the decoder function
                    has to take a string argument
    :type decoder: :class:`collections.Callable`

    .. versionchanged:: 0.2.0
       The ``default`` option becomes to accept only callable objects.
       Below 0.2.0, ``default`` is not a function but a value which
       is simply used as it is.

    """

    #: (:class:`str`) The XML attribute name.
    name = None

    #: (:class:`str`) The optional XML namespace URI.
    xmlns = None

    #: (:class:`tuple`) The pair of (:attr:`xmlns`, :attr:`name`).
    key_pair = None

    #: (:class:`bool`) Whether it is required for the element.
    required = None

    #: (:class:`collections.Callable`) The function that returns default
    #: value when the attribute is not present.  The function takes an
    #: argument which is an :class:`Element` instance.
    #:
    #: .. versionchanged:: 0.2.0
    #:    It becomes to accept only callable objects.  Below 0.2.0,
    #:    :attr:`default` attribute is not a function but a value which
    #:    is simply used as it is.
    default = None

    def __init__(self, name, codec=None, xmlns=None, required=False,
                 default=None, encoder=None, decoder=None):
        if not (default is None or callable(default)):
            raise TypeError('default must be callable, not ' + repr(default))
        super(Attribute, self).__init__(codec=codec,
                                        encoder=encoder,
                                        decoder=decoder)
        self.name = name
        self.xmlns = xmlns
        self.key_pair = xmlns, name
        self.required = bool(required)
        self.default = default

    def __get__(self, obj, cls=None):
        if isinstance(obj, Element):
            attrs = obj._attrs
            if self.default is None:
                return attrs.get(self)
            return attrs.setdefault(self, self.default(obj))
        return self

    def __set__(self, obj, value):
        if isinstance(obj, Element):
            obj._attrs[self] = value


class Content(CodecDescriptor):
    """Declare possible text nodes as a descriptor.

    :param codec: an optional codec object to use.  if it's callable and
                  not an instance of :class:`Codec`, its return value will
                  be used instead.  it means this can take class object of
                  :class:`Codec` subtype that is not instantiated yet
                  unless the constructor require any arguments
    :type codec: :class:`Codec`, :class:`collections.Callable`
    :param encoder: an optional function that encodes Python value into
                    XML text value e.g. :func:`str()`.  the encoder function
                    has to take an argument
    :type encoder: :class:`collections.Callable`
    :param decoder: an optional function that decodes XML text value into
                    Python value e.g. :func:`int()`.  the decoder function
                    has to take a string argument
    :type decoder: :class:`collections.Callable`

    """

    def __get__(self, obj, cls=None):
        if isinstance(obj, Element):
            if getattr(obj, '_root', None):
                root = obj._root()
                if getattr(root, '_parser', None):
                    handler = root._handler
                    while (obj._content is None and
                           (not handler.stack or handler.stack[-1])):
                        if not root._parse_next():
                            break
            return obj._content
        return self

    def __set__(self, obj, value):
        obj._content = value

    def read(self, element, value):
        """Read raw ``value`` from XML, decode it, and then set the attribute
        for content of the given ``element`` to the decoded value.

        .. note::

           Internal method.

        """
        decoded = self.decode(value, element)
        self.__set__(element, decoded)


class Element(object):
    """Represent an element in XML document.

    It provides the default constructor which takes keywords
    and initializes the attributes by given keyword arguments.
    For example, the following code that uses the default
    constructor::

        assert issubclass(Person, Element)

        author = Person(
            name='Hong Minhee',
            url='http://dahlia.kr/'
        )

    is equivalent to the following code::

        author = Person()
        author.name = 'Hong Minhee'
        author.url = 'http://dahlia.kr/'

    """

    __slots__ = ('_attrs', '_content', '_data', '_parent', '_root', '_partial',
                 '_hints')

    def __init__(self, _parent=None, **attributes):
        self._attrs = getattr(self, '_attrs', {})  # FIXME
        self._content = getattr(self, '_content', None)
        self._data = getattr(self, '_data', {})
        # _partial has three states:
        # 0. the element is completely loaded
        # 1. the element is partially loaded
        # 2. the element is partially loaded, but _hints are loaded
        self._partial = 0
        self._hints = {}
        if _parent is not None:
            if not isinstance(_parent, Element):
                raise TypeError('expected a {0.__module__}.{0.__name__} '
                                'instance, not {1!r}'.format(Element, _parent))
            self._parent = weakref.ref(_parent)
            self._root = _parent._root
            if hasattr(self._root(), '_handler'):
                self._stack_top = (1 if self._root() is self
                                   else len(self._root()._handler.stack))
                self._partial = 1
        cls = type(self)
        acceptable_desc_types = Descriptor, Content, Attribute, property
        # FIXME: ^-- hardcoded type list
        for attr_name, attr_value in attributes.items():
            if isinstance(getattr(cls, attr_name, None), acceptable_desc_types):
                setattr(self, attr_name, attr_value)
            else:
                raise SchemaError('{0.__module__}.{0.__name__} has no such '
                                  'attribute: {1}'.format(cls, attr_name))

    def __entity_id__(self):
        """Identify the entity object.  It returns the entity object itself
        by default, but should be overridden.

        :returns: any value to identify the entity object

        """
        return self

    def __merge_entities__(self, other):
        """Merge two entities (``self`` and ``other``).  It can return one
        of the two, or even a new entity object.  This method is used by
        :class:`~libearth.session.Session` objects to merge conflicts between
        concurrent updates.

        :param other: other entity to merge.  it's guaranteed that it's
                      older session's (note that it doesn't mean this entity
                      is older than ``self``, but the session's last update
                      is)
        :type other: :class:`Element`
        :returns: on of the two, or even an new entity object that merges
                  two entities
        :rtype: :class:`Element`

        .. note::

           The default implementation simply returns ``self``.
           That means the entity of the newer session will always win
           unless the method is overridden.

        """
        return self

    @classmethod
    def __coerce_from__(cls, value):
        """Cast a value which isn't an instance of the element type to
        the element type.  It's useful when a boxed element type could
        be more naturally represented using builtin type.

        For example, :class:`~libearth.feed.Mark` could be represented
        as a boolean value, and :class:`~libearth.feed.Text` also
        could be represented as a string.

        The following example shows how the element type can be
        automatically casted from string by implementing
        :meth:`__coerce_from__()` class method::

            @classmethod
            def __coerce_from__(cls, value):
                if isinstance(value, str):
                    return Text(value=value)
                raise TypeError('expected a string or Text')

        """
        raise TypeError('expected an instance of {0.__module__}.{0.__name__}, '
                        'not {1!r}'.format(cls, value))


class DocumentElement(Element):
    """The root element of the document.

    .. attribute:: __tag__

       (:class:`str`) Every :class:`DocumentElement` subtype has to define
       this attribute to the root tag name.

    .. attribute:: __xmlns__

       (:class:`str`) A :class:`DocumentElement` subtype may define this
       attribute to the XML namespace of the document element.

    """

    __slots__ = '_parser', '_iterator', '_handler'

    __tag__ = NotImplemented
    __xmlns__ = None

    def __init__(self, _parent=None, **kwargs):
        cls = type(self)
        if cls.__tag__ is NotImplemented:
            raise NotImplementedError(
                '{0.__module__}.{0.__name__}.__tag__ is not defined; '
                'every subtype of {1.__module__}.{1.__name__} has to '
                'define __tag__ attribute'.format(cls, DocumentElement)
            )
        elif not isinstance(cls.__tag__, string_type):
            raise TypeError(
                '__tag__ has to be a string, not ' + repr(cls.__tag__)
            )
        self._root = weakref.ref(self)
        super(DocumentElement, self).__init__(_parent or self, **kwargs)

    def _parse_next(self):
        """Parse the next step of iteration.

        :returns: whether it's not complete or consumed all.
                  :const:`False` if it's completely consumed
        :rtype: :class:`bool`

        """
        parser = getattr(self, '_parser', None)
        if not parser:
            return False
        if isinstance(parser, PullReader):
            if parser.feed():
                return True
        else:
            try:
                chunk = next(self._iterator)
            except StopIteration:
                pass
            else:
                parser.feed(chunk)
                return True
        try:
            parser.close()
        except xml.sax.SAXException:
            pass
        return False


class ElementList(collections.MutableSequence):
    """List-like object to represent multiple chidren.  It makes the parser
    to lazily consume the buffer when an element of a particular offset
    is requested.

    You can extend methods or properties for a particular element type
    using :func:`element_list_for()` class decorator e.g.::

        @element_list_for(Link)
        class LinkList(collections.Sequence):
            '''Specialized ElementList for Link elements.'''

            def filter_by_mimetype(self, mimetype):
                '''Filter links by their mimetype.'''
                return [link for link in self if link.mimetype == mimetype]

    Extended methods/properties can be used for element lists for the type::

        assert isinstance(feed.links, LinkList)
        assert isinstance(feed.links, ElementList)
        feed.links.filter_by_mimetype('text/html')

    """

    __slots__ = ('element', 'descriptor', 'value_type',
                 'tag', 'xmlns', 'key_pair')

    #: (:class:`collections.MutableMapping`) The internal table for
    #: specialized subtypes used by :meth:`register_specialized_type()`
    #: method and :func:`element_list_for()` class decorator.
    specialized_types = {}

    @classmethod
    def register_specialized_type(cls, value_type, specialized_type):
        """Register specialized :class:`collections.Sequence` type for
        a particular ``value_type``.

        An imperative version of :func`element_list_for()` class decorator.

        :param value_type: a particular element type that ``specialized_type``
                             would be used for instead of default
                             :class:`ElementList` class.
                             it has to be a subtype of :class:`Element`
        :type value_type: :class:`type`
        :param specialized_type: a :class:`collections.Sequence` type which
                                 extends methods and properties for
                                 ``value_type``
        :type specialized_type: :class:`type`

        """
        if not isinstance(value_type, type):
            raise TypeError('value_type must be a type object, not ' +
                            repr(value_type))
        elif not issubclass(value_type, Element):
            raise TypeError(
                'value_type must be a subtype of {0.__module__}.{0.__name__},'
                ' not {1.__module__}.{1.__name__}'.format(Element, value_type)
            )
        elif not isinstance(specialized_type, type):
            raise TypeError('specialized_type must be a type object, not ' +
                            repr(specialized_type))
        elif not issubclass(specialized_type, collections.Sequence):
            raise TypeError(
                'specialized_type must be a subtype of {0.__module__}.'
                '{0.__name__}, not {1.__module__}.{1.__name__}'
                ''.format(collections.Sequence, specialized_type)
            )
        try:
            t, t2 = cls.specialized_types[value_type]
        except KeyError:
            pass
        else:
            if t is specialized_type or t2 is specialized_type:
                return
            raise TypeError(
                '{0.__module__}.{0.__name__} already has its own specialized '
                'subtype: {1.__module__}.{1.__name__}'.format(value_type, t)
            )
        cls.specialized_types[value_type] = specialized_type, None

    def __new__(cls, element, descriptor, value_type=None):
        try:
            supcls, subcls = cls.specialized_types[value_type]
        except KeyError:
            subcls = cls
        else:
            if issubclass(supcls, cls):
                cls = supcls
                subcls = supcls
                cls.specialized_types[value_type] = supcls, supcls
            if subcls is None:
                subcls = type(supcls.__name__, (cls, supcls), {})
                subcls.__module__ = supcls.__module__
                cls.specialized_types[value_type] = supcls, subcls
        return object.__new__(subcls)

    def __init__(self, element, descriptor, value_type=None):
        if not isinstance(element, Element):
            raise TypeError(
                'element must be an instance of {0.__module__}.{0.__name__}, '
                'not {1!r}'.format(Element, element)
            )
        elif not isinstance(descriptor, Descriptor):
            raise TypeError(
                'descriptor must be an instance of {0.__module__}.{0.__name__}'
                ', not {1!r}'.format(Descriptor, descriptor)
            )
        elif not (value_type is None or isinstance(value_type, type)):
            raise TypeError('value_type must be a type, not ' +
                            repr(value_type))
        self.element = element
        self.descriptor = descriptor
        self.value_type = value_type

    def consume_buffer(self):
        """Consume the buffer for the parser.  It returns a generator,
        so can be stopped using :keyword:`break` statement by caller.

        .. note::

           Internal method.

        """
        root_ref = getattr(self.element, '_root', None)
        if not root_ref:
            return
        root = root_ref()
        if not getattr(root, '_parser', None):
            return
        data = self.element._data
        while not self.consumes_all():
            yield data
            if not root._parse_next():
                break
        yield data

    def consumes_all(self):
        element = self.element
        if getattr(element, '_parent', None) is None:
            return True
        parent = element._parent()
        root = element._root()
        handler = getattr(root, '_handler', None)
        if handler is None:
            return True
        stack = handler.stack
        top = element._stack_top
        if len(stack) < top:
            return True
        for context in reversed(stack):
            if context.reserved_value is parent:
                return False
        return True

    def consume_index(self, index, ignore_length_hint=False):
        if isinstance(index, slice):
            if index.start is not None and index.start >= 0 and \
               index.stop is not None and index.stop >= 0:
                index = max(index.start, index.stop)
            else:
                index = -1
        key = self.descriptor
        if index >= 0:
            if not ignore_length_hint:
                length_hint = self._length_hint
                if length_hint is not None and index >= length_hint:
                    raise IndexError('list index out of range')
            for data in self.consume_buffer():
                if key in data and len(data[key]) > index:
                    return data[key]
        else:
            for data in self.consume_buffer():
                continue
        return self.element._data.setdefault(key, [])

    def validate_value(self, value):
        if self.value_type is None or isinstance(value, self.value_type):
            return value
        elif self.value_type is string_type:
            raise TypeError('expected a string, not ' + repr(value))
        elif issubclass(self.value_type, Element):
            value = self.value_type.__coerce_from__(value)
            if isinstance(value, self.value_type):
                return value
        raise TypeError('expected an instance of {0.__module__}.{0.__name__}, '
                        'not {1!r}'.format(self.value_type, value))

    @property
    def _length_hint(self):
        element = self.element
        if element._partial == 1:
            for data in self.consume_buffer():
                if element._partial != 1:
                    break
        try:
            length_hint = element._hints[self.descriptor]['length']
        except KeyError:
            return
        return int(length_hint)

    @_length_hint.setter
    def _length_hint(self, value):
        hints = self.element._hints.setdefault(self.descriptor, {})
        hints['length'] = str(value)

    def __len__(self):
        length_hint = self._length_hint
        if length_hint is not None:
            return length_hint
        key = self.descriptor
        data = self.element._data
        for data in self.consume_buffer():
            continue
        try:
            lst = data[key]
        except KeyError:
            length = 0
        else:
            length = len(lst)
        self._length_hint = length
        return length

    def __getitem__(self, index):
        return self.consume_index(index)[index]

    def __setitem__(self, index, value):
        data = self.consume_index(index)
        if isinstance(index, slice):
            data[index] = map(self.validate_value, value)
        else:
            data[index] = self.validate_value(value)

    def __delitem__(self, index):
        data = self.consume_index(index)
        del data[index]
        self._length_hint = len(data)

    def insert(self, index, value):
        data = self.consume_index(index, ignore_length_hint=True)
        data.insert(index, self.validate_value(value))
        self._length_hint = len(data)

    def __nonzero__(self):
        length_hint = self._length_hint
        if length_hint is not None:
            return bool(length_hint)
        data = self.consume_index(1)
        return bool(data)

    def __repr__(self):
        consumed = self.element._data.get(self.descriptor, [])
        list_repr = repr(consumed)
        if not self.consumes_all():
            if consumed:
                list_repr = list_repr[:-1] + ', ...]'
            else:
                list_repr = '[...]'
        return '<{0.__module__}.{0.__name__} {1}>'.format(
            type(self), list_repr
        )


class element_list_for(object):
    """Class decorator which registers specialized :class:`ElementList`
    subclass for a particular ``value_type`` e.g.::

        @element_list_for(Link)
        class LinkList(collections.Sequence):
            '''Specialized ElementList for Link elements.'''

            def filter_by_mimetype(self, mimetype):
                '''Filter links by their mimetype.'''
                return [link for link in self if link.mimetype == mimetype]

    :param value_type: a particular element type that ``specialized_type``
                         would be used for instead of default
                         :class:`ElementList` class.
                         it has to be a subtype of :class:`Element`
    :type value_type: :class:`type`

    """

    def __init__(self, value_type):
        if not isinstance(value_type, type):
            raise TypeError('value_type must be a type object, not ' +
                            repr(value_type))
        elif not issubclass(value_type, Element):
            raise TypeError(
                'value_type must be a subtype of {0.__module__}.{0.__name__},'
                ' not {1.__module__}.{1.__name__}'.format(Element, value_type)
            )
        self.value_type = value_type

    def __call__(self, specialized_type):
        ElementList.register_specialized_type(self.value_type, specialized_type)
        return specialized_type


# Semi-structured record type for only internal use.
ParserContext = collections.namedtuple(
    'ParserContext',
    'tag xmlns descriptor reserved_value content_buffer'
)


class ContentHandler(xml.sax.handler.ContentHandler):
    """Event handler implementation for SAX parser.

    It maintains the stack that contains parsing contexts of
    what element is lastly open, what descriptor is associated
    to the element, and the buffer for chunks of content characters
    the element has.  Every context is represented as the namedtuple
    :class:`ParserContext`.

    Each time its events (:meth:`startElement()`, :meth:`characters()`,
    and :meth:`endElement()`) are called, it forwards the data to
    the associated descriptor.  :class:`Descriptor` subtypes
    implement :meth:`~Descriptor.start_element()` method and
    :meth:`~Descriptor.end_element()`.

    """

    def __init__(self, document):
        self.document = weakref.ref(document)
        self.stack = []

    def load_hint(self, parent_element, tag, attrs):
        xmlns, name = tag
        if not (xmlns == SCHEMA_XMLNS and name == 'hint'):
            parent_element._partial = 2
            return False
        child_tags = inspect_child_tags(type(parent_element))
        child_xmlns = attrs.get((None, 'tag-xmlns'))
        child_name = attrs[None, 'tag']
        attr, desc = child_tags[child_xmlns, child_name]
        hint_dict = parent_element._hints.setdefault(desc, {})
        hint_dict[attrs[None, 'id']] = attrs[None, 'value']
        return True

    def startElementNS(self, tag, qname, attrs):
        xmlns, name = tag
        try:
            parent_context = self.stack[-1]
        except IndexError:
            # document element
            doc = self.document()
            if self.load_hint(doc, tag, attrs):
                return
            expected = getattr(doc, '__xmlns__', None), doc.__tag__
            if tag != expected:
                raise IntegrityError('document element must be {0}, '
                                     'not {1}'.format(expected, name))
            self.stack.append(
                ParserContext(
                    tag=name,
                    xmlns=xmlns,
                    descriptor=None,
                    reserved_value=doc,
                    content_buffer=[]
                )
            )
            reserved_value = doc
        else:
            parent_element = parent_context.reserved_value
            if self.load_hint(parent_element, tag, attrs):
                return
            element_type = type(parent_element)
            child_tags = inspect_child_tags(element_type)
            try:
                attr, child = child_tags[tag]
            except KeyError:
                available_children = [
                    '{0} (namespace: {1})'.format(name, ns) if xmlns else name
                    for ns, name in child_tags
                ]
                available_children.sort()
                available_children = ', '.join(available_children)
                if xmlns:
                    raise IntegrityError(
                        'unexpected element: {0} (namespace: {1}); available '
                        'elements: {2}'.format(name, xmlns, available_children)
                    )
                raise IntegrityError(
                    'unexpected element: {0}; available elements: '.format(
                        name,
                        available_children
                    )
                )
            if isinstance(child, Descriptor):
                reserved_value = child.start_element(parent_element, attr)
                self.stack.append(
                    ParserContext(
                        tag=name,
                        xmlns=xmlns,
                        descriptor=child,
                        reserved_value=reserved_value,
                        content_buffer=[]
                    )
                )
            else:
                if xmlns:
                    raise IntegrityError('unexpected element: {0} (namespace: '
                                         '{1})'.format(name, xmlns))
                raise IntegrityError('unexpected element: ' + name)
        if isinstance(reserved_value, Element):
            instance = reserved_value
            instance_type = type(instance)
            attributes = inspect_attributes(instance_type)
            instance_attrs_dict = instance._attrs
            for xml_attr, raw_value in attrs.items():
                try:
                    _, attr_desc = attributes[xml_attr]
                except KeyError:
                    continue
                instance_attrs_dict[attr_desc] = attr_desc.decode(
                    raw_value,
                    instance
                )

    def characters(self, content):
        context = self.stack[-1]
        context.content_buffer.append(content)

    def endElementNS(self, tag, qname):
        xmlns, name = tag
        if xmlns == SCHEMA_XMLNS:
            return
        context = self.stack.pop()
        assert name == context.tag
        assert xmlns == context.xmlns
        text = ''.join(context.content_buffer)
        if context.descriptor is None:
            # context.reserved_value is root document
            attr = inspect_content_tag(type(context.reserved_value))
            if attr is not None:
                content_desc = attr[1]
                content_desc.read(context.reserved_value, text)
            context.reserved_value._partial = 0
        else:
            context.descriptor.end_element(context.reserved_value, text)


def complete(element):
    """Completely load the given ``element``.

    :param element: an element loaded by :func:`read()`
    :type element: :class:`Element`

    """
    if not isinstance(element, Element):
        raise TypeError('element must be an instance of {0.__module__}.'
                        '{0.__name__}, not {1!r}'.format(Element, element))
    if element._partial:
        parse_next = element._root()._parse_next
        while element._partial:
            parse_next()


def is_partially_loaded(element):
    """Return whether the given ``element`` is not completely loaded
    by :func:`read()` yet.

    :param element: an element
    :type element: :class:`Element`
    :returns: whether :const:`True` if the given ``element`` is partially
              loaded
    :rtype: :class:`bool`

    """
    if not isinstance(element, Element):
        raise TypeError('element must be an instance of {0.__module__}.'
                        '{0.__name__}, not {1!r}'.format(Element, element))
    return bool(element._partial)


def index_descriptors(element_type):
    """Index descriptors of the given ``element_type`` to make them
    easy to be looked up by their identifiers (pairs of XML namespace URI
    and tag name).

    :param element_type: a subtype of :class:`Element`
                         to index its descriptors
    :type element_type: :class:`type`

    .. note::

       Internal function.

    """
    if not (isinstance(element_type, type) and
            issubclass(element_type, Element)):
        raise TypeError('element_type must be a subtype of {0.__name__}.'
                        '{0.__name__}, not {1!r}'.format(Element, element_type))
    elif (issubclass(element_type, DocumentElement) and
          getattr(element_type, '__xmlns__', None)):
        xmlns_set = set([element_type.__xmlns__])
    else:
        xmlns_set = set()
    attributes = {}
    child_tags = {}
    content = None
    element_type.__xmlns_set__ = xmlns_set  # to avoid infinite loop
    for attr in dir(element_type):
        desc = getattr(element_type, attr)
        if isinstance(desc, Content):
            if content is not None:
                raise DescriptorConflictError(
                    'there are more than a descriptor for the element content '
                    '(text node): {0!r} and {1!r}; there must not be any '
                    'duplicate descriptors for the same content'.format(
                        content[0], attr
                    )
                )
            content = attr, desc
        elif isinstance(desc, Attribute):
            if desc.key_pair in attributes:
                if desc.xmlns:
                    name = '{{{0}}}{1}'.format(*desc.key_pair)
                else:
                    name = desc.name
                raise DescriptorConflictError(
                    'there are more than a descriptor for the same attribute '
                    '{0!r}: {1!r} and {2!r}; there must not be any duplicate '
                    'descriptors for the same attribute'.format(
                        name, attributes[desc.key_pair], attr
                    )
                )
            attributes[desc.key_pair] = attr, desc
            if desc.xmlns:
                xmlns_set.add(desc.xmlns)
        elif isinstance(desc, Descriptor):
            if desc.key_pair in child_tags:
                if desc.xmlns:
                    tag = '{{{0}}}{1}'.format(*desc.key_pair)
                else:
                    tag = desc.tag
                raise DescriptorConflictError(
                    'there are more than a descriptor for the same element '
                    '{0!r}: {1!r} and {2!r}; there must not be any duplicate '
                    'descriptors for the same element'.format(
                        tag, child_tags[desc.key_pair], attr
                    )
                )
            child_tags[desc.key_pair] = attr, desc
            if desc.xmlns:
                xmlns_set.add(desc.xmlns)
            if isinstance(desc, Child):  # FIXME: should be polymorphic
                xmlns_set.update(inspect_xmlns_set(desc.element_type))
    element_type.__xmlns_set__ = frozenset(xmlns_set)
    element_type.__attributes__ = attributes
    element_type.__child_tags__ = child_tags
    element_type.__content_tag__ = content


def inspect_xmlns_set(element_type):
    """Get the set of XML namespaces used in the given ``element_type``,
    recursively including all child elements.

    :param element_type: a subtype of :class:`Element` to inspect
    :type element_type: :class:`type`
    :returns: a set of uri strings of used all xml namespaces
    :rtype: :class:`collections.Set`

    .. note::

       Internal function.

    """
    try:
        return element_type.__xmlns_set__
    except AttributeError:
        index_descriptors(element_type)
        return element_type.__xmlns_set__


def inspect_attributes(element_type):
    """Get the dictionary of :class:`Attribute` descriptors of
    the given ``element_type``.

    :param element_type: a subtype of :class:`Element` to inspect
    :type element_type: :class:`type`
    :returns: a dictionary of attribute identifiers (pairs of
              xml namespace uri and xml attribute name) to pairs of
              instance attribute name and associated :class:`Attribute`
              descriptor
    :rtype: :class:`collections.Mapping`

    .. note::

       Internal function.

    """
    try:
        return element_type.__attributes__
    except AttributeError:
        index_descriptors(element_type)
        return element_type.__attributes__


def inspect_child_tags(element_type):
    """Get the dictionary of :class:`Descriptor` objects of
    the given ``element_type``.

    :param element_type: a subtype of :class:`Element` to inspect
    :type element_type: :class:`type`
    :returns: a dictionary of child node identifiers (pairs of
              xml namespace uri and tag name) to pairs of
              instance attribute name and associated :class:`Descriptor`
    :rtype: :class:`collections.Mapping`

    .. note::

       Internal function.

    """
    try:
        child_tags = element_type.__child_tags__
    except AttributeError:
        index_descriptors(element_type)
        child_tags = element_type.__child_tags__
    else:
        # FIXME: it should be tested, and considered in inspect_content_tag(),
        # inspect_xmlns_set(), and inspect_attributes() as well.
        if any(hasattr(sup, '__child_tags__') and
               sup.__child_tags__ is child_tags
               for sup in element_type.__bases__):
            index_descriptors(element_type)
            child_tags = element_type.__child_tags__
    return child_tags


def inspect_content_tag(element_type):
    """Gets the :class:`Content` descriptor of the given ``element_type``.

    :param element_type: a subtype of :class:`Element` to inspect
    :type element_type: :class:`type`
    :returns: a pair of instance attribute name and associated
              :class:`Content` descriptor
    :rtype: :class:`tuple`

    .. note::

       Internal function.

    """
    try:
        content = element_type.__content_tag__
    except AttributeError:
        index_descriptors(element_type)
        content = element_type.__content_tag__
    return content


#: (:class:`collections.Sequence`) The list of :mod:`xml.sax` parser
#: implementations to try to import.
PARSER_LIST = []

if platform.python_implementation() == 'IronPython':
    PARSER_LIST = ['libearth.compat.clrxmlreader']


def read(cls, iterable):
    """Initialize a document in read mode by opening the ``iterable``
    of XML string.  ::

        with open('doc.xml', 'rb') as f:
            read(Person, f)

    Returned document element is not fully read but partially loaded
    into memory, and then lazily (and eventually) loaded when these
    are actually needed.

    :param cls: a subtype of :class:`DocumentElement`
    :type cls: :class:`type`
    :param iterable: chunks of XML string to read
    :type iterable: :class:`collections.Iterable`
    :returns: initialized document element in read mode
    :rtype: :class:`DocumentElement`

    """
    if not isinstance(cls, type):
        raise TypeError('cls must be a type object, not ' + repr(cls))
    elif not issubclass(cls, DocumentElement):
        raise TypeError(
            'cls must be a subtype of {0.__module__}.{0.__name__}, not '
            '{1.__module__}.{1.__name__}'.format(cls, DocumentElement)
        )
    doc = cls()
    parser = xml.sax.make_parser(PARSER_LIST)
    handler = ContentHandler(doc)
    parser.setContentHandler(handler)
    parser.setFeature(xml.sax.handler.feature_namespaces, True)
    if isinstance(parser, PullReader):
        parser.prepareParser(iterable)
    else:
        doc._iterator = iter(iterable)
    doc._parser = parser
    doc._handler = handler
    stack = handler.stack
    while not stack:
        if not doc._parse_next():
            break
    Element.__init__(doc, doc)
    return doc


def validate(element, recurse=True, raise_error=True):
    """Validate the given ``element`` according to the schema.  ::

        from libearth.schema import IntegrityError, validate

        try:
            validate(element)
        except IntegrityError:
            print('the element {0!r} is invalid!'.format(element))

    :param element: the element object to validate
    :type element: :class:`Element`
    :param recurse: recursively validate the whole tree (child nodes).
                    :const:`True` by default
    :type recurse: :class:`bool`
    :param raise_error: raise exception when the ``element`` is invalid.
                        if it's :const:`False` it returns :const:`False`
                        instead of raising an exception.
                        :const:`True` by default
    :type raise_error: :class:`bool`
    :returns: :const:`True` if the ``element`` is valid.
              :const:`False` if the ``element`` is invalid and
              ``raise_error`` option is :const:`False``
    :raise IntegrityError: when the ``element`` is invalid and
                           ``raise_error`` option is :const:`True`

    """
    element_type = type(element)
    for name, desc in inspect_attributes(element_type).values():
        if desc.required and not getattr(element, name, None):
            if raise_error:
                raise IntegrityError(
                    '{0.__module__}.{0.__name__}.{1} is required, but '
                    '{2!r} lacks it'.format(element_type, name, element)
                )
            return False
    for name, desc in inspect_child_tags(element_type).values():
        children = getattr(element, name, None)
        if not desc.multiple:
            children = children,
        for child_element in children:
            if desc.required and not child_element:
                if raise_error:
                    raise IntegrityError(
                        '{0.__module__}.{0.__name__}.{1} is required, but '
                        '{2!r} lacks it'.format(element_type, name, element)
                    )
                return False
            if recurse and child_element is not None and \
               isinstance(desc, Child):
                if validate(child_element,
                            recurse=True,
                            raise_error=raise_error):
                    continue
                return False
    return True


class write(collections.Iterable):
    r"""Write the given ``document`` to XML string.  The return value is
    an iterator that yields chunks of an XML string.  ::

        with open('doc.xml', 'w') as f:
            for chunk in write(document):
                f.write(chunk)

    :param document: the document element to serialize
    :type document: :class:`DocumentElement`
    :param validate: whether validate the ``document`` or not.
                     :const:`True` by default
    :type validate: :class:`bool`
    :param indent: an optional string to be used for indent.
                   default is four spaces (``'    '``)
    :type indent: :class:`str`
    :param newline: an optional character to be used for newline.
                    default is ``'\n'``
    :type newline: :class:`str`
    :param canonical_order: make the order of attributes and child nodes
                            consistent to any python versions and
                            implementations.  useful for testing.
                            :const:`False` by default
    :type canonical_order: :class:`bool`
    :param hints: export hint values as well.  hints improves efficiency of
                  :func:`read()`.  :const:`True` by default
    :type hints: :class:`bool`
    :param as_bytes: return chunks as :class:`bytes` (:class:`str` in Python 2)
                     if :const:`True`.  return chunks as :class:`str`
                     (:class:`unicode` in Python 3) if :const:`False`.
                     return chunks as default string type (:class:`str`)
                     by default
    :returns: chunks of an XML string
    :rtype: :class:`collections.Iterable`

    """

    def __init__(self, document, validate=True, indent='  ', newline='\n',
                 canonical_order=False, hints=True, as_bytes=None):
        if not isinstance(document, DocumentElement):
            raise TypeError(
                'document must be an instance of {0.__module__}.{0.__name__}, '
                'not {1!r}'.format(DocumentElement, document)
            )
        self.document = document
        self.document_type = type(document)
        self.validate = validate
        self.indent = indent
        self.newline = newline
        self.as_bytes = as_bytes
        self.sort = sorted if canonical_order else lambda l, *a, **k: l
        self.hints = hints
        xmlns_set = inspect_xmlns_set(self.document_type)
        self.xmlns_alias = dict(
            (uri, 'ns{0}'.format(i))
            for i, uri in enumerate(self.sort(xmlns_set))
        )
        if hints:
            self.xmlns_alias[SCHEMA_XMLNS] = 'libearth'

    def __iter__(self):
        result = itertools.chain(['<?xml version="1.0" encoding="utf-8"?>\n'],
                                 self.export(self.document,
                                             self.document_type.__tag__,
                                             self.document_type.__xmlns__))
        if UNICODE_BY_DEFAULT and self.as_bytes:
            return (binary_type(chunk, 'utf-8') for chunk in result)
        elif not UNICODE_BY_DEFAULT and self.as_bytes is False:
            return (chunk.decode('utf-8') for chunk in result)
        return result

    if UNICODE_BY_DEFAULT:
        encode = staticmethod(lambda s: s)
    else:
        encode = staticmethod(lambda s: s.encode('utf-8'))

    def export(self, element, tag, xmlns, depth=0):
        if self.validate:
            validate(element, recurse=False, raise_error=True)
        element_type = type(element)
        for s in itertools.repeat(self.indent, depth):
            yield s
        yield '<'
        if xmlns:
            yield self.xmlns_alias[xmlns]
            yield ':'
        yield tag
        quoteattr = xml.sax.saxutils.quoteattr
        if not depth:
            for uri, prefix in self.sort(self.xmlns_alias.items(),
                                         key=operator.itemgetter(0)):
                yield ' xmlns:'
                yield prefix
                yield '='
                yield quoteattr(uri)
        attr_descriptors = self.sort(
            inspect_attributes(element_type).values(),
            key=operator.itemgetter(0)
        )
        encode = self.encode
        for attr, desc in attr_descriptors:
            raw_attr_value = getattr(element, attr, None)
            if raw_attr_value is None:
                continue
            encoded_attr_value = desc.encode(raw_attr_value, element)
            if encoded_attr_value is None:
                continue
            elif not isinstance(encoded_attr_value, string_type):
                raise EncodeError(
                    '{0.__module__}.{0.__name__}.{1} attribute value {2!r} '
                    'is incorrectly encoded to {3!r}'.format(
                        element_type, attr, raw_attr_value, encoded_attr_value
                    )
                )
            yield ' '
            if desc.xmlns:
                yield self.xmlns_alias[desc.xmlns]
                yield ':'
            yield desc.name
            yield '='
            quoted_attr = quoteattr(encoded_attr_value)
            if not isinstance(quoteattr, binary_type):
                quoted_attr = encode(quoted_attr)
            yield quoted_attr
        content = inspect_content_tag(element_type)
        children = inspect_child_tags(element_type)
        escape = xml.sax.saxutils.escape
        newline = self.newline
        if content or children:
            assert not (content and children)
            yield '>'
            if content:
                raw_content_value = getattr(element, content[0], None)
                encoded_content_value = content[1].encode(raw_content_value,
                                                          element)
                if encoded_content_value is not None:
                    if not isinstance(encoded_content_value, string_type):
                        raise EncodeError(
                            '{0.__module__}.{0.__name__}.{1} attribute value '
                            '{2!r} is incorrectly encoded to {3!r}'.format(
                                element_type, content[0],
                                raw_content_value, encoded_content_value
                            )
                        )
                    yield encode(escape(encoded_content_value))
            else:
                if self.hints:
                    hints = self.sort(
                        (desc.tag, desc.xmlns, hint_dict)
                        for desc, hint_dict in element._hints.items()
                    )
                    for hint_tag, hint_xmlns, hint_dict in hints:
                        for hint_id, hint_val in self.sort(hint_dict.items()):
                            yield newline
                            for s in itertools.repeat(self.indent, depth + 1):
                                yield s
                            yield '<'
                            yield self.xmlns_alias[SCHEMA_XMLNS]
                            yield ':hint tag='
                            yield quoteattr(hint_tag)
                            if hint_xmlns:
                                yield ' tag-xmlns='
                                yield quoteattr(hint_xmlns)
                            yield ' id='
                            if not isinstance(hint_id, binary_type):
                                hint_id = encode(hint_id)
                            yield quoteattr(hint_id)
                            yield ' value='
                            if not isinstance(hint_val, binary_type):
                                hint_val = encode(hint_val)
                            yield quoteattr(hint_val)
                            yield '/>'
                children = self.sort(
                    children.values(),
                    key=lambda pair: pair[1].descriptor_counter
                )
                for attr, desc in children:
                    child_elements = getattr(element, attr, None)
                    if not desc.multiple:
                        child_elements = [child_elements]
                    if desc.sort_key is not None:
                        child_elements = sorted(
                            child_elements,
                            key=desc.sort_key,
                            reverse=bool(desc.sort_reverse)
                        )
                    for child_element in child_elements:
                        if isinstance(desc, Text):  # FIXME: remove type query
                            if child_element is None:
                                continue
                            encoded_child = desc.encode(child_element, element)
                            if encoded_child is None:
                                continue
                            elif not isinstance(encoded_child, string_type):
                                raise EncodeError(
                                    '{0.__module__}.{0.__name__}.{1} attribute '
                                    'value {2!r} is incorrectly encoded to '
                                    '{3!r}'.format(element_type,
                                                   attr,
                                                   child_element,
                                                   encoded_child)
                                )
                            yield newline
                            for s in itertools.repeat(self.indent, depth + 1):
                                yield s
                            yield '<'
                            if desc.xmlns:
                                yield self.xmlns_alias[desc.xmlns]
                                yield ':'
                            yield desc.tag
                            yield '>'
                            yield encode(escape(encoded_child))
                            yield '</'
                            if desc.xmlns:
                                yield self.xmlns_alias[desc.xmlns]
                                yield ':'
                            yield desc.tag
                            yield '>'
                        elif child_element is not None:
                            yield newline
                            subiter = self.export(child_element,
                                                  desc.tag,
                                                  desc.xmlns,
                                                  depth=depth + 1)
                            for chunk in subiter:
                                yield chunk
                yield newline
            if children:
                for s in itertools.repeat(self.indent, depth):
                    yield s
            yield '</'
            if xmlns:
                yield self.xmlns_alias[xmlns]
                yield ':'
            yield tag
            yield '>'
        else:
            yield '/>'



""":mod:`libearth.session` --- Isolate data from other installations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides merging facilities to avoid conflict between concurrent
updates of the same document/entity from different devices (installations).
There are several concepts here.

:class:`Session` abstracts installations on devices.  For example, if you
have a laptop, a tablet, and a mobile phone, and two apps are installed on
the laptop, then there have to be four sessions: `laptop-1`, `laptop-2`,
`table-1`, and `phone-1`.  You can think of it as branch if you are familiar
with DVCS.

:class:`Revision` abstracts timestamps of updated time.  An important thing
is that it preserves its session as well.

Base revisions (:attr:`MergeableDocumentElement.__base_revisions__`) show
what revisions the current revision is built on top of.  In other words,
what revisions were merged into the current revision.  :class:`RevisionSet`
is a dictionary-like data structure to represent them.

"""
import collections
import datetime
import re
import uuid
import xml.sax

from .codecs import Rfc3339
from .compat import string_type
from .compat.xmlpullreader import PullReader
from .schema import (PARSER_LIST, Attribute, Codec, DecodeError,
                     DocumentElement, Element, EncodeError, inspect_attributes,
                     inspect_child_tags, inspect_content_tag)
from .tz import now

__all__ = ('SESSION_XMLNS', 'MergeableDocumentElement', 'Revision',
           'RevisionCodec', 'RevisionParserHandler', 'RevisionSet',
           'RevisionSetCodec', 'Session',
           'ensure_revision_pair', 'parse_revision')


#: (:class:`str`) The XML namespace name used for session metadata.
SESSION_XMLNS = 'http://earthreader.org/session/'


class Session(object):
    """The unit of device (more abstractly, *installation*) that updates
    the same document (e.g. :class:`~libearth.feed.Feed`).  Every session
    must have its own unique :attr:`identifier` to avoid conflict between
    concurrent updates from different sessions.

    :param identifier: the unique identifier.  automatically generated
                       using :mod:`uuid` if not present
    :type identifier: :class:`str`

    """

    #: (:class:`re.RegexObject`) The regular expression pattern that matches
    #: to allowed identifiers.
    IDENTIFIER_PATTERN = re.compile(r'^[-a-z0-9_.]+$', re.IGNORECASE)

    #: (:class:`collections.MutableMapping`) The pool of interned sessions.
    #: It's for maintaining single sessions for the same identifiers.
    interns = {}

    #: (:class:`str`) The session identifier.  It has to be distinguishable
    #: from other devices/apps, but consistent for the same device/app.
    identifier = None

    def __new__(cls, identifier=None):
        if not (identifier is None or isinstance(identifier, string_type)):
            raise TypeError('identifier must be a string, not ' +
                            repr(identifier))
        identifier = str(identifier) if identifier else str(uuid.getnode())
        if not cls.IDENTIFIER_PATTERN.match(identifier):
            raise ValueError('invalid identifier format: ' + repr(identifier))
        try:
            session = cls.interns[identifier]
        except KeyError:
            session = super(Session, cls).__new__(cls)
            session.identifier = identifier
            cls.interns[identifier] = session
        return session

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return hash(self.identifier)

    def revise(self, document):
        """Mark the given ``document`` as the latest revision of the current
        session.

        :param document: mergeable document to mark
        :type document: :class:`MergeableDocumentElement`

        """
        if not isinstance(document, MergeableDocumentElement):
            raise TypeError(
                'document must be an instance of {0.__module__}.{0.__name__}, '
                'not {1!r}'.format(MergeableDocumentElement, document)
            )
        if document.__base_revisions__:
            updated_at = max(now(), max(document.__base_revisions__.values()))
        else:
            updated_at = now()
        document.__revision__ = Revision(self, updated_at)

    def pull(self, document):
        """Pull the ``document`` (of possibly other session) to the current
        session.

        :param document: the document to pull from the possibly other session
                         to the current session
        :type document: :class:`MergeableDocumentElement`
        :returns: the clone of the given ``document`` with the replaced
                  :attr:`~MergeableDocumentElement.__revision__`.
                  note that the :attr:`Revision.updated_at` value won't
                  be revised.  it could be the same object to the given
                  ``document`` object if the session is the same
        :rtype: :class:`MergeableDocumentElement`

        """
        if not isinstance(document, MergeableDocumentElement):
            raise TypeError(
                'expected a {0.__module__}.{0.__name__} instance, not '
                '{1!r}'.format(MergeableDocumentElement, document)
            )
        rev = document.__revision__
        if rev is not None and rev.session is self:
            return document
        # TODO: It could be efficiently implemented using SAX parser with
        #       less memory use.
        element_type = type(document)
        copy = element_type()
        for name, desc in inspect_child_tags(element_type).values():
            if desc.multiple:
                value = list(getattr(document, name, []))
            else:
                value = getattr(document, name, None)
            setattr(copy, name, value)
        for name, _ in inspect_attributes(element_type).values():
            setattr(copy, name, getattr(document, name, None))
        content = inspect_content_tag(element_type)
        if content is not None:
            name = content[0]
            setattr(copy, name, getattr(document, name))
        if rev:
            copy.__revision__ = Revision(self, rev.updated_at)
        else:
            self.revise(copy)
        return copy

    def merge(self, a, b, force=False):
        """Merge the given two documents and return new merged document.
        The given documents are not manipulated in place.  Two documents
        must have the same type.

        :param a: the first document to be merged
        :type a: :class:`MergeableDocumentElement`
        :param b: the second document to be merged
        :type b: :class:`MergeableDocumentElement`
        :param force: by default (:const:`False`) it doesn't merge but
                      simply pull a or b if one already contains other.
                      if ``force`` is :const:`True` it always merge
                      two.  it assumes ``b`` is newer than ``a``

        """
        element_type = type(a)
        cls_b = type(b)
        if element_type is not cls_b:
            raise TypeError(
                'two document must have the same type; but {0.__module__}.'
                '{0.__name__} and {1.__module__}.{1.__name__} are not the '
                'same type'.format(element_type, cls_b)
            )
        if not force:
            if a.__base_revisions__.contains(b.__revision__):
                return self.pull(a)
            elif b.__base_revisions__.contains(a.__revision__):
                return self.pull(b)
        # The latest one should be `b`.
        if not force and a.__revision__.updated_at > b.__revision__.updated_at:
            a, b = b, a
        merged = a.__merge_entities__(b)
        self.revise(merged)
        merged_revisions = a.__base_revisions__.merge(
            b.__base_revisions__,
            RevisionSet([a.__revision__, b.__revision__])
        )
        merged.__base_revisions__ = merged_revisions
        self.revise(merged)
        return merged

    def __str__(self):
        return self.identifier

    def __repr__(self):
        return '{0.__module__}.{0.__name__}({1!r})'.format(type(self),
                                                           self.identifier)


#: The named tuple type of (:class:`Session`, :class:`datetime.datetime`) pair.
Revision = collections.namedtuple('Revision', 'session updated_at')


def ensure_revision_pair(pair, force_cast=False):
    """Check the type of the given ``pair`` and error unless it's a valid
    revision pair (:class:`Session`, :class:`datetime.datetime`).

    :param pair: a value to check
    :type pair: :class:`collections.Sequence`
    :param force_cast: whether to return the casted value to :class:`Revision`
                       named tuple type
    :type force_cast: :class:`bool`
    :returns: the revision pair
    :rtype: :class:`Revision`, :class:`collections.Sequence`

    """
    try:
        session, updated_at = pair
    except ValueError:
        raise TypeError('expected a pair, not ' + repr(pair))
    if not isinstance(session, Session):
        raise TypeError('{0!r} is not an instance of {1.__module__}.'
                        '{1.__name__}'.format(session, Session))
    elif not isinstance(updated_at, datetime.datetime):
        raise TypeError('{0!r} is not an instance of {1.__module__}.{1.'
                        '__name__}'.format(updated_at, datetime.datetime))
    if isinstance(pair, Revision) or not force_cast:
        return pair
    return Revision(*pair)


class RevisionSet(collections.Mapping):
    """Set of :class:`Revision` pairs.  It provides dictionary-like
    mapping protocol.

    :param revisions: the iterable of
                      (:class:`Session`, :class:`datetime.datetime`) pairs
    :type revisions: :class:`collections.Iterable`

    """

    def __init__(self, revisions=[]):
        self.revisions = dict(map(ensure_revision_pair, revisions))

    def __len__(self):
        return len(self.revisions)

    def __iter__(self):
        return iter(self.revisions)

    def __getitem__(self, session):
        return self.revisions[session]

    def items(self):
        """The list of (:class:`Session`, :class:`datetime.datetime`) pairs.

        :return: the list of :class:`Revision` instances
        :rtype: :class:`collections.ItemsView`

        """
        return [Revision(*pair) for pair in super(RevisionSet, self).items()]

    def copy(self):
        """Make a copy of the set.

        :returns: a new equivalent set
        :rtype: :class:`RevisionSet`

        """
        return type(self)(self.items())

    def merge(self, *sets):
        """Merge two or more :class:`RevisionSet`\ s.  The latest time
        remains for the same session.

        :param \*sets: one or more :class:`RevisionSet` objects to merge
        :returns: the merged set
        :rtype: :class:`RevisionSet`

        """
        cls = type(self)
        if not sets:
            raise TypeError('expected one or more {0.__module__}.{0.__name__} '
                            'objects'.format(cls))
        sets = sets + (self,)
        sessions = set()
        for revisions in sets:
            if not isinstance(revisions, cls):
                raise TypeError('{0!r} is not an instance of {1.__module__}.'
                                '{1.__name__}'.format(revisions, cls))
            sessions.update(revisions)
        return cls(
            Revision(session, max(s[session] for s in sets if session in s))
            for session in sessions
        )

    def contains(self, revision):
        """Find whether the given ``revision`` is already merged to
        the revision set.  In other words, return :const:`True`
        if the ``revision`` doesn't have to be merged to the revision set
        anymore.

        :param revision: the revision to find whether it has to be merged
                         or not
        :type revision: :class:`Revision`
        :returns: :const:`True` if the ``revision`` is included in
                  the revision set, or :const:`False`
        :rtype: :class:`bool`

        """
        try:
            return self[revision.session] >= revision.updated_at
        except KeyError:
            return False

    def __repr__(self):
        return '{0.__module__}.{0.__name__}({1!r})'.format(type(self),
                                                           self.items())


class RevisionCodec(Codec):
    """Codec to encode/decode :class:`Revision` pairs.

    >>> from libearth.tz import utc
    >>> session = Session('test-identifier')
    >>> updated_at = datetime.datetime(2013, 9, 22, 3, 43, 40, tzinfo=utc)
    >>> rev = Revision(session, updated_at)
    >>> RevisionCodec().encode(rev)
    'test-identifier 2013-09-22T03:43:40Z'

    """

    #: (:class:`Rfc3339`) The internally used codec to encode
    #: :attr:`Revision.updated_at` time to :rfc:`3339` format.
    RFC3339_CODEC = Rfc3339(prefer_utc=True)

    def encode(self, value):
        try:
            session, updated_at = ensure_revision_pair(value)
        except TypeError as e:
            raise EncodeError(str(e))
        return '{0} {1}'.format(session, self.RFC3339_CODEC.encode(updated_at))

    def decode(self, text):
        try:
            identifier, updated_at = text.split()
        except ValueError:
            raise DecodeError(repr(text) + ' is an invalid revision format')
        try:
            session = Session(identifier)
        except ValueError:
            raise DecodeError(repr(identifier) +
                              ' is and invalid session identifier')
        return Revision(
            session=session,
            updated_at=self.RFC3339_CODEC.decode(updated_at)
        )


class RevisionSetCodec(RevisionCodec):
    r"""Codec to encode/decode multiple :class:`Revision` pairs.

    >>> from datetime import datetime
    >>> from libearth.tz import utc
    >>> revs = RevisionSet([
    ...     (Session('a'), datetime(2013, 9, 22, 16, 58, 57, tzinfo=utc)),
    ...     (Session('b'), datetime(2013, 9, 22, 16, 59, 30, tzinfo=utc)),
    ...     (Session('c'), datetime(2013, 9, 22, 17, 0, 30, tzinfo=utc))
    ... ])
    >>> encoded = RevisionSetCodec().encode(revs)
    >>> encoded
    'c 2013-09-22T17:00:30Z,\nb 2013-09-22T16:59:30Z,\na 2013-09-22T16:58:57Z'
    >>> RevisionSetCodec().decode(encoded)
    libearth.session.RevisionSet([
        Revision(session=libearth.session.Session('b'),
                 updated_at=datetime.datetime(2013, 9, 22, 16, 59, 30,
                                              tzinfo=libearth.tz.Utc())),
        Revision(session=libearth.session.Session('c'),
                 updated_at=datetime.datetime(2013, 9, 22, 17, 0, 30,
                                              tzinfo=libearth.tz.Utc())),
        Revision(session=libearth.session.Session('a'),
                 updated_at=datetime.datetime(2013, 9, 22, 16, 58, 57,
                                              tzinfo=libearth.tz.Utc()))
    ])

    """

    #: (:class:`re.RegexObject`) The regular expression pattern that matches
    #: to separator substrings between revision pairs.
    SEPARATOR_PATTERN = re.compile(r'\s*,\s*')

    def encode(self, value):
        if not isinstance(value, RevisionSet):
            raise EncodeError('{0!r} is not an instance of {1.__module__}.'
                              '{1.__name__}'.format(value, RevisionSet))
        encode_pair = super(RevisionSetCodec, self).encode
        pairs = value.items()
        if not isinstance(pairs, list):
            pairs = list(pairs)
        pairs.sort(key=lambda pair: pair[1], reverse=True)
        return ',\n'.join(map(encode_pair, pairs))

    def decode(self, text):
        decode_pair = super(RevisionSetCodec, self).decode
        pairs = self.SEPARATOR_PATTERN.split(text) if text.strip() else []
        return RevisionSet(map(decode_pair, pairs))


class MergeableDocumentElement(DocumentElement):
    """Document element which is mergeable using :class:`Session`."""

    #: (:class:`Revision`) The revision of the document.
    __revision__ = Attribute('revision', RevisionCodec, xmlns=SESSION_XMLNS)

    #: (:class:`RevisionSet`) The set of revisions that its current
    #: :attr:`revision` is built on top of.  That means these revisions
    #: no longer need to be merged.
    __base_revisions__ = Attribute('bases', RevisionSetCodec,
                                   xmlns=SESSION_XMLNS,
                                   default=lambda _: RevisionSet())

    def __merge_entities__(self, other):
        entity_id = lambda e: (e.__entity_id__()
                               if isinstance(e, Element)
                               else e)
        element_type = type(self)
        merged = element_type()
        for attr_name, desc in inspect_child_tags(element_type).values():
            if desc.multiple:
                a_list = getattr(self, attr_name, [])
                identifiers = dict((entity_id(entity), entity)
                                   for entity in a_list)
                merged_attr = list(a_list)
                for element in getattr(other, attr_name, []):
                    eid = entity_id(element)
                    try:
                        entity = identifiers[eid]
                    except KeyError:
                        merged_element = element
                    else:
                        merged_attr.remove(entity)
                        if isinstance(element, Element):
                            merged_element = element.__merge_entities__(entity)
                        else:
                            merged_element = element
                    identifiers[eid] = merged_element
                    merged_attr.append(merged_element)
            else:
                older_attr = getattr(self, attr_name, None)
                newer_attr = getattr(other, attr_name, None)
                if older_attr is None:
                    merged_attr = newer_attr
                elif newer_attr is None:
                    merged_attr = older_attr
                elif isinstance(newer_attr, Element):
                    merged_attr = newer_attr.__merge_entities__(older_attr)
                else:
                    merged_attr = newer_attr
            setattr(merged, attr_name, merged_attr)
        for attr_name, _ in inspect_attributes(element_type).values():
            setattr(merged, attr_name,
                    getattr(other, attr_name, getattr(self, attr_name, None)))
        content = inspect_content_tag(element_type)
        if content is not None:
            name = content[0]
            setattr(merged, name,
                    getattr(other, name, getattr(self, name, None)))
        return merged


class RevisionParserHandler(xml.sax.handler.ContentHandler):
    """SAX content handler that picks session metadata
    (:attr:`~MergeableDocumentElement.__revision__` and
    :attr:`~MergeableDocumentElement.__base_revisions__`) from the given
    document element.

    Parsed result goes :attr:`revision` and :attr:`base_revisions`.

    Used by :func:`parse_revision()`.

    """

    #: (:class:`bool`) Represents whether the parsing is complete.
    done = None

    #: (:class:`Revision`) The parsed
    #: :attr:`~MergeableDocumentElement.__revision__`.  It might be
    #: :const:`None`.
    revision = None

    #: (:class:`RevisionSet`). The parsed
    #: :attr:`~MergeableDocumentElement.__base_revisions__`.  It might be
    #: :const:`None`.

    def __init__(self):
        self.done = False
        self.revision = None
        self.base_revisions = None

    def startElementNS(self, tag, qname, attrs):
        if self.done:
            return
        revision_desc = MergeableDocumentElement.__revision__
        bases_desc = MergeableDocumentElement.__base_revisions__
        self.revision = attrs.get((revision_desc.xmlns, revision_desc.name))
        self.base_revisions = attrs.get((bases_desc.xmlns, bases_desc.name))
        self.done = True


def parse_revision(iterable):
    """Efficiently parse only :attr:`~MergeableDocumentElement.__revision__`
    and :attr:`~MergeableDocumentElement.__base_revisions__` from the given
    ``iterable`` which contains chunks of XML.  It reads only head of
    the given document, and ``iterable`` will be not completely consumed
    in most cases.

    Note that it doesn't validate the document.

    :param iterable: chunks of bytes which contains
                     a :class:`MergeableDocumentElement` element
    :type iterable: :class:`collections.Iterable`
    :returns: a pair of (:attr:`~MergeableDocumentElement.__revision__`,
              :attr:`~MergeableDocumentElement.__base_revisions__`).
              it might be :const:`None` if the document is not stamped
    :rtype: :class:`collections.Sequence`

    """
    parser = xml.sax.make_parser(PARSER_LIST)
    handler = RevisionParserHandler()
    parser.setContentHandler(handler)
    parser.setFeature(xml.sax.handler.feature_namespaces, True)
    if isinstance(parser, PullReader):
        parser.prepareParser(iterable)
        while not handler.done and parser.feed():
            pass
    else:
        iterator = iter(iterable)
        while not handler.done:
            try:
                chunk = next(iterator)
            except StopIteration:
                break
            parser.feed(chunk)
    if handler.revision is None:
        return
    rev_codec = RevisionCodec()
    revset_codec = RevisionSetCodec()
    return (rev_codec.decode(handler.revision),
            revset_codec.decode(handler.base_revisions))



""":mod:`libearth.crawler` --- Crawler
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Crawl feeds.

"""
import collections
import functools
import logging
import re
import sys

try:
    import urllib.request as urllib2
except ImportError:
    import urllib2
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

from .compat.parallel import parallel_map
from .feed import Link
from .parser.autodiscovery import AutoDiscovery, get_format
from .subscribe import SubscriptionSet
from .version import VERSION


__all__ = 'DEFAULT_TIMEOUT', 'CrawlError', 'CrawlResult', 'crawl', 'get_feed'


#: (:class:`numbers.Integral`) The default timeout for connection attempts.
#: 10 seconds.
#:
#: .. versionadded:: 0.3.0
DEFAULT_TIMEOUT = 10


def open_url(url, *args, **kwargs):
    if isinstance(url, Request):
        request = url
    else:
        request = urllib2.Request(url)
    request.add_header('User-agent', '{0}/{1}'.format(__package__, VERSION))
    return urllib2.urlopen(request, *args, **kwargs)


def crawl(feed_urls, pool_size, timeout=DEFAULT_TIMEOUT):
    """Crawl feeds in feed list using thread.

    :param feed_urls: feed urls to crawl
    :type feed_urls: :class: `collections.Sequence`
    :param pool_size: the number of concurrent workers
    :type pool_size: :class:`numbers.Integral`
    :param timeout: optional timeout for connection attempts.
                    :const:`DEFAULT_TIMEOUT` is used if omitted
    :type timeout: :class:`numbers.Integral`
    :returns: a set of :class:`CrawlResult` objects
    :rtype: :class:`collections.Iterable`

    .. versionchanged:: 0.3.0

       It became to return a set of :class:`CrawlResult`\ s instead of
       :class:`tuple`\ s.

    .. versionchanged:: 0.3.0

       The parameter ``feeds`` was renamed to ``feed_urls``.

    .. versionadded:: 0.3.0

       Added optional ``timeout`` parameter.

    """
    if type(timeout) is type(DEFAULT_TIMEOUT) and timeout == DEFAULT_TIMEOUT:
        func = get_feed
    else:
        func = functools.partial(get_feed, timeout=int(timeout))
    return parallel_map(pool_size, func, feed_urls)


def get_feed(feed_url, timeout=DEFAULT_TIMEOUT):
    # TODO: should be documented
    logger = logging.getLogger(__name__ + '.get_feed')
    try:
        f = open_url(feed_url, timeout=timeout)
        feed_xml = f.read()
        f.close()
        parser = get_format(feed_xml)
        if parser is None:
            logger.warn('failed to detect the format of %s', feed_url)
            logger.debug('the response body of %s:\n%s', feed_url, feed_xml)
            raise CrawlError(feed_url,
                             'failed to detect the format of ' + feed_url)
        feed, crawler_hints = parser(feed_xml, feed_url)
        self_uri = None
        for link in feed.links:
            if link.relation == 'self':
                self_uri = link.uri
        if not self_uri:
            feed.links.append(Link(relation='self', uri=feed_url,
                                   mimetype=f.info()['content-type']))
        feed.entries = sorted(feed.entries, key=lambda entry: entry.updated_at,
                              reverse=True)
        favicon = feed.links.favicon
        if favicon is None:
            permalink = feed.links.permalink
            if permalink:
                try:
                    f = open_url(permalink.uri, timeout=timeout)
                except IOError:
                    pass
                else:
                    content_type = f.headers['content-type']
                    html = f.read()
                    f.close()
                    if isinstance(html, bytes) and \
                       not isinstance(html, str):
                        match = re.search(r';\s*charset\s*=\s*([^;\s]+)',
                                          content_type)
                        enc = match.group(1) if match else 'utf-8'
                        html = html.decode(enc, 'replace')
                    _, icon_urls = AutoDiscovery().find(html)
                    if icon_urls:
                        favicon = urlparse.urljoin(permalink.uri,
                                                   icon_urls[0])
                if favicon is None:
                    favicon = urlparse.urljoin(permalink.uri, '/favicon.ico')
                    req = Request(favicon, method='HEAD')
                    try:
                        f = open_url(req, timeout=timeout)
                    except (IOError, OSError):
                        favicon = None
                    else:
                        if f.getcode() != 200:
                            favicon = None
                        f.close()
        else:
            favicon = favicon.uri
        return CrawlResult(feed_url, feed, crawler_hints, favicon)
    except Exception as e:
        logger.exception(
            '%s: %s', feed_url, e
        )
        raise CrawlError(feed_url, '{0} failed: {1}'.format(feed_url, e))


class CrawlResult(collections.Sequence):
    """The result of each crawl of a feed.

    It mimics triple of (:attr:`url`, :attr:`feed`, :attr:`hints`) for
    backward compatibility to below 0.3.0, so you can still take these
    values using tuple unpacking, though it's not recommended way to
    get these values anymore.

    .. versionadded:: 0.3.0

    """

    #: (:class:`str`) The crawled :attr:`feed` url.
    url = None

    #: (:class:`~libearth.feed.Feed`) The crawled feed.
    feed = None

    #: (:class:`collections.Mapping`) The extra hints for the crawler
    #: e.g. ``skipHours``, ``skipMinutes``, ``skipDays``.
    #: It might be :const:`None`.
    hints = None

    #: (:class:`str`) The favicon url of the :attr:`feed` if exists.
    #: It might be :const:`None`.
    icon_url = None

    def __init__(self, url, feed, hints, icon_url=None):
        self.url = url
        self.feed = feed
        self.hints = hints
        self.icon_url = icon_url

    def add_as_subscription(self, subscription_set):
        """Add it as a subscription to the given ``subscription_set``.

        :param subscription_set: a subscription list or category to add
                                 a new subscription
        :type subscription_set: :class:`~libearth.subscribe.SubscriptionSet`
        :returns: the created subscription object
        :rtype: :class:`~libearth.subscribe.Subscription`

        """
        if not isinstance(subscription_set, SubscriptionSet):
            raise TypeError(
                'expected an instance of {0.__module__}.{0.__name__}, '
                'not {1!r}'.format(SubscriptionSet, subscription_set)
            )
        return subscription_set.subscribe(self.feed, icon_uri=self.icon_url)

    def __len__(self):
        return 3

    def __getitem__(self, index):
        if index == 0 or index == -3:
            return self.url
        elif index == 1 or index == -2:
            return self.feed
        elif index == 2 or index == -1:
            return self.hints
        raise IndexError('index out of range')


class CrawlError(IOError):
    """Error which rises when crawling given url failed.

    .. versionadded:: 0.3.0
       Added ``feed_uri`` parameter and corresponding :attr:`feed_uri`
       attribute.

    """

    #: (:class:`str`) The errored feed uri.
    feed_uri = None

    def __init__(self, feed_uri, *args, **kwargs):
        super(CrawlError, self).__init__(*args, **kwargs)
        self.feed_uri = feed_uri


if sys.version_info >= (3, 3):
    # Since Python 3.3 urllib.request.Request can take a method argument
    Request = urllib2.Request
else:
    class Request(urllib2.Request):
        """Request which can take a ``method`` argument."""

        def __init__(self, *args, **kwargs):
            method = kwargs.pop('method', None)
            urllib2.Request.__init__(self, *args, **kwargs)
            self._method = method

        def get_method(self):
            return self._method or urllib2.Request.get_method(self)



""":mod:`libearth.parser.rss2` --- RSS 2.0 parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Parsing RSS 2.0 feed.

"""
import datetime
import email.utils
import logging
import re

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

from ..codecs import Rfc3339, Rfc822
from ..compat import IRON_PYTHON
from ..compat.etree import fromstring
from ..feed import (Category, Content, Entry, Feed, Generator, Link,
                    Person, Text)
from ..schema import DecodeError
from ..tz import FixedOffset, guess_tzinfo_by_locale, now, utc
from .atom import ATOM_XMLNS_SET
from .util import normalize_xml_encoding


GUID_PATTERN = re.compile('^(\{{0,1}([0-9a-fA-F]){8}-([0-9a-fA-F]){4}-([0-9'
                          'a-fA-F]){4}-([0-9a-fA-F]){4}-([0-9a-fA-F]){12}\}'
                          '{0,1})$')
CONTENT_XMLNS = '{http://purl.org/rss/1.0/modules/content/}'


def parse_rss(xml, feed_url=None, parse_entry=True):
    """Parse RSS 2.0 XML and translate it into Atom.

    To make the feed data valid in Atom format, ``id`` and ``link[rel=self]``
    fields would become the url of the feed.

    If ``pubDate`` is not present, ``updated`` field will be from
    the latest entry's ``updated`` time, or the time it's crawled instead.

    :param xml: rss 2.0 xml string to parse
    :type xml: :class:`str`
    :param parse_item: whether to parse items (entries) as well.
                       it's useful when to ignore items when retrieve
                       ``<source>``.  :const:`True` by default
    :type parse_item: :class:`bool`
    :returns: a pair of (:class:`~libearth.feed.Feed`, crawler hint)
    :rtype: :class:`tuple`

    """
    root = fromstring(normalize_xml_encoding(xml))
    channel = root.find('channel')
    items = channel.findall('item')
    default_tzinfo = guess_default_tzinfo(root, feed_url)
    feed_data, crawler_hints = rss_get_channel_data(
        channel,
        feed_url,
        default_tzinfo
    )
    if parse_entry:
        feed_data.entries = rss_get_item_data(items, default_tzinfo)
    check_valid_as_atom(feed_data)
    return feed_data, crawler_hints


def check_valid_as_atom(feed_data):
    # FIXME: It doesn't only "check" the feed_data but manipulates it
    # if not valid.  I think the function should be renamed.
    if feed_data.updated_at is None:
        if feed_data.entries:
            try:
                feed_data.updated_at = max(entry.updated_at
                                           for entry in feed_data.entries
                                           if entry.updated_at)
            except ValueError:
                feed_data.updated_at = now()
        else:
            feed_data.updated_at = now()
    if feed_data.title is None:
        feed_data.title = feed_data.subtitle
        # FIXME: what should we do when there's even no subtitle?
    for entry in feed_data.entries:
        if entry.updated_at is None:
            entry.updated_at = feed_data.updated_at


def rss_get_channel_data(root, feed_url, default_tzinfo):
    _log = logging.getLogger(__name__ + '.rss_get_channel_data')
    feed_data = Feed(id=feed_url)
    crawler_hints = {}
    contributors = []
    for data in root:
        if data.tag == 'title':
            feed_data.title = Text(value=data.text or '')
        elif data.tag == 'description':
            feed_data.subtitle = Text(type='text', value=data.text)
        elif (any(data.tag == '{' + ns + '}link' for ns in ATOM_XMLNS_SET) and
              data.get('href')):
            # FIXME: duplicate with rss_get_item_data()
            link = Link(uri=data.get('href'),
                        relation=data.get('rel', 'alternate'),
                        mimetype=data.get('type'))
            feed_data.links.append(link)
        elif not data.text:
            _log.warn('Empty tag: %s', data)
            continue
        elif data.tag == 'link':
            link = Link(uri=data.text,
                        relation='alternate',
                        mimetype='text/html')
            feed_data.links.append(link)
        elif data.tag == 'copyright':
            feed_data.rights = Text(value=data.text)
        elif data.tag in ('managingEditor', 'webMaster'):
            contributors.extend(parse_person(data.text, True))
            feed_data.contributors = contributors
        elif data.tag == 'pubDate':
            feed_data.updated_at = parse_datetime(data.text, default_tzinfo)
        elif data.tag == 'category':
            feed_data.categories.append(parse_category(data))
        elif data.tag == 'generator':
            generator = None
            try:
                if urlparse.urlparse(data.text).scheme in ('http', 'https'):
                    generator = Generator(uri=data.text)
            except ValueError:
                pass
            feed_data.generator = generator or Generator(value=data.text)
        elif data.tag == 'lastBuildDate':
            crawler_hints['lastBuildDate'] = parse_datetime(data.text,
                                                            default_tzinfo)
        elif data.tag == 'ttl':
            crawler_hints['ttl'] = data.text
        elif data.tag == 'skipHours':
            crawler_hints['skipHours'] = data.text
        elif data.tag == 'skipMinutes':
            crawler_hints['skipMinutes'] = data.text
        elif data.tag == 'skipDays':
            crawler_hints['skipDays'] = data.text
        elif data.tag != 'item':
            _log.warn('Unknown tag: %s', data)
    if all(l.relation != 'self' for l in feed_data.links):
        feed_data.links.insert(0, Link(relation='self', uri=feed_url))
    return feed_data, crawler_hints


def rss_get_item_data(entries, default_tzinfo):
    _log = logging.getLogger(__name__ + '.rss_get_item_data')
    entries_data = []
    for entry in entries:
        entry_data = Entry()
        for data in entry:
            if data.tag == 'enclosure':
                link = Link(
                    relation='enclosure',
                    mimetype=data.get('type'),
                    uri=data.get('url')
                )
                entry_data.links.append(link)
            elif data.tag == 'source':
                from ..crawler import open_url
                from .autodiscovery import get_format
                url = data.get('url')
                f = open_url(url)  # FIXME: propagate timeout option
                xml = f.read()
                parser = get_format(xml)
                source, _ = parser(xml, url, parse_entry=False)
                entry_data.source = source
            elif any(data.tag == '{' + ns + '}link'
                     for ns in ATOM_XMLNS_SET) and data.get('href'):
                # FIXME: duplicate with rss_get_channel_data()
                link = Link(uri=data.get('href'),
                            relation=data.get('rel', 'alternate'),
                            mimetype=data.get('type'))
                entry_data.links.append(link)
            elif not data.text:
                _log.warn('Empty tag: %s', data)
                continue
            elif data.tag == 'title':
                entry_data.title = Text(value=data.text)
            elif data.tag == 'link':
                link = Link(uri=data.text,
                            relation='alternate',
                            mimetype='text/html')
                entry_data.links.append(link)
            elif data.tag == 'comments':
                link = Link(uri=data.text, relation='discussion')
                entry_data.links.append(link)
            elif data.tag == 'description' and not entry_data.content:
                entry_data.content = Content(type='html', value=data.text)
            elif data.tag == CONTENT_XMLNS + 'encoded':
                entry_data.content = Content(type='html', value=data.text)
            elif data.tag == 'author':
                entry_data.authors = parse_person(data.text, True)
            elif data.tag == 'category':
                entry_data.categories.append(parse_category(data))
            elif data.tag == 'guid':
                isPermalink = data.get('isPermalink')
                if data.text.startswith('http://') and isPermalink != 'False':
                    entry_data.id = data.text
                elif GUID_PATTERN.match(data.text):
                    entry_data.id = 'urn:uuid:' + data.text
            elif data.tag == 'pubDate':
                entry_data.published_at = parse_datetime(data.text,
                                                         default_tzinfo)
                # TODO 'pubDate' is optional in RSS 2, but 'updated' in Atom
                #       is required element, so we have to fill some value to
                #       entry.updated_at.
            else:
                _log.warn('Unknown tag: %s', data)
        if entry_data.updated_at is None:
            entry_data.updated_at = entry_data.published_at
        if entry_data.id is None:
            entry_data.id = entry_data.links[0].uri \
                if entry_data.links else ''
        entries_data.append(entry_data)
    return entries_data


_rfc3339 = Rfc3339()
_rfc822 = Rfc822()
_datetime_formats = [
    ('%Y-%m-%d %H:%M:%S', None),  # daumwebtoon
    ('%m/%d/%Y %H:%M:%S GMT', utc),  # msdn
    ('%m/%d/%y %H:%M:%S GMT', utc),  # msdn
    ('%a, %d %b %Y %H:%M:%S GMT 00:00:00 GMT', utc),  # msdn
    ('%Y.%m.%d %H:%M:%S', None),  # imbcnews
    ('%d %b %Y %H:%M:%S %z', None),  # lee-seungjae
]


def parse_datetime(string, default_tzinfo):
    # https://github.com/earthreader/libearth/issues/30
    try:
        return _rfc822.decode(string)
    except DecodeError:
        pass
    try:
        return _rfc3339.decode(string)
    except DecodeError:
        pass
    for fmt, tzinfo in _datetime_formats:
        try:
            if IRON_PYTHON:
                # IronPython strptime() seems to ignore whitespace
                string = string.replace(' ', '|')
                fmt = fmt.replace(' ', '|')
            if fmt.endswith('%z'):
                dt = datetime.datetime.strptime(string[:-5], fmt[:-2])
                tz_sign = -1 if string[-5:-4] == '-' else 1
                tz_hour = int(string[-4:-2])
                tz_min = int(string[-2:])
                tzinfo = FixedOffset(tz_sign * (tz_hour * 60 + tz_min))
            else:
                dt = datetime.datetime.strptime(string, fmt)
            return dt.replace(tzinfo=tzinfo or default_tzinfo)
        except ValueError:
            continue
    raise ValueError('failed to parse datetime: ' + repr(string))


def parse_person(string, as_list=False):
    name, email_addr = email.utils.parseaddr(string)
    if '@' not in email_addr:
        if not name:
            name = email_addr
        email_addr = None
    if not name:
        name = email_addr
    if not name:
        return [] if as_list else None
    person = Person(name=name, email=email_addr or None)
    return [person] if as_list else person


def parse_category(element):
    return Category(
        term=element.text,
        scheme_uri=element.attrib.get('domain')
    )


def guess_default_tzinfo(root, url):
    """Guess what time zone is implied in the feed by seeing the TLD of
    the ``url`` and its ``<language>`` tag.

    """
    lang = root.find('channel/language')
    if lang is None or not lang.text:
        return utc
    lang = lang.text.strip()
    if len(lang) == 5 and lang[2] == '-':
        lang = lang[:2]
    parsed = urlparse.urlparse(url)
    domain = parsed.hostname.rsplit('.', 1)
    country = domain[1] if len(domain) > 1 and len(domain[1]) == 2 else None
    return guess_tzinfo_by_locale(lang, country) or utc



""":mod:`libearth.compat.xmlpullreader` --- Pulling SAX parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import xml.sax.xmlreader

__all__ = 'PullReader',


class PullReader(xml.sax.xmlreader.XMLReader):
    """SAX parser interface which provides similar but slightly less power
    than :class:`~xml.sax.xmlreader.IncremenetalParser`.

    :class:`~xml.sax.xmlreader.IncrementalParser` can feed arbitrary length
    of bytes while it can't determine how long bytes to feed.

    """

    def feed(self):
        """This method makes the parser to parse the next step node,
        emitting the corresponding events.

        :meth:`feed()` may raise :exc:`~xml.sax.SAXException`.

        :returns: whether the stream buffer is not empty yet
        :rtype: :class:`bool`
        :raises xml.sax.SAXException: when something goes wrong

        """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement feed() method'.format(PullReader)
        )

    def prepareParser(self, iterable):
        """This method is called by the parse implementation to allow
        the SAX 2.0 driver to prepare itself for parsing.

        :param iterable: iterable of :class:`bytes`
        :type iterable: :class:`collections.Iterable`

        """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement prepareParser() method'.format(PullReader)
        )

    def close(self):
        """This method is called when the entire XML document has been
        passed to the parser through the feed method, to notify the
        parser that there are no more data. This allows the parser to
        do the final checks on the document and empty the internal
        data buffer.

        The parser will not be ready to parse another document until
        the reset method has been called.

        :meth:`close()` may raise :exc:`~xml.sax.SAXException`.

        :raises xml.sax.SAXException: when something goes wrong

        """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement close() method'.format(PullReader)
        )

    def reset(self):
        """This method is called after close has been called to reset
        the parser so that it is ready to parse new documents.
        The results of calling parse or feed after close without calling
        reset are undefined.

        """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement reset() method'.format(PullReader)
        )



""":mod:`libearth.sanitizer` --- Sanitize HTML tags
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import cgi
try:
    import htmlentitydefs
    import HTMLParser
except ImportError:
    from html import entities as htmlentitydefs, parser as HTMLParser
import re

from .compat import unichr

__all__ = 'HtmlSanitizer', 'MarkupTagCleaner', 'clean_html', 'sanitize_html'


def clean_html(html):
    """Strip *all* markup tags from ``html`` string.
    That means, it simply makes the given ``html`` document a plain text.

    :param html: html string to clean
    :type html: :class:`str`
    :returns: cleaned plain text
    :rtype: :class:`str`

    """
    parser = MarkupTagCleaner()
    parser.feed(html)
    return ''.join(parser.fed)


def sanitize_html(html):
    """Sanitize the given ``html`` string.  It removes the following
    tags and attributes that are not secure nor useful for RSS reader layout:

    - ``<script>`` tags
    - ``display: none;`` styles
    - JavaScript event attributes e.g. ``onclick``, ``onload``
    - ``href`` attributes that start with ``javascript:``, ``jscript:``,
      ``livescript:``, ``vbscript:``, ``data:``, ``about:``, or ``mocha:``.

    :param html: html string to sanitize
    :type html: :class:`str`
    :returns: cleaned plain text
    :rtype: :class:`str`

    """
    parser = HtmlSanitizer()
    parser.feed(html)
    return ''.join(parser.fed)


class MarkupTagCleaner(HTMLParser.HTMLParser):
    """HTML parser that is internally used by :func:`clean_html()` function."""

    entity_map = htmlentitydefs.name2codepoint

    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def handle_entityref(self, name):
        try:
            codepoint = self.entity_map[name]
        except KeyError:
            pass
        else:
            self.fed.append(unichr(codepoint))

    def handle_charref(self, name):
        if name.startswith('x'):
            codepoint = int(name[1:], 16)
        else:
            codepoint = int(name)
        self.fed.append(unichr(codepoint))


class HtmlSanitizer(HTMLParser.HTMLParser):
    """HTML parser that is internally used by :func:`sanitize_html()`
    function.

    """

    #: (:class:`re.RegexObject`) The regular expression pattern that matches to
    #: disallowed CSS properties.
    DISALLOWED_STYLE_PATTERN = re.compile(
        r'(^|;)\s*display\s*:\s*[a-z-]+\s*(?:;\s*|$)',
        re.IGNORECASE
    )

    #: (:class:`collections.Set`) The set of disallowed URI schemes e.g.
    #: ``javascript:``.
    DISALLOWED_SCHEMES = frozenset([
        'javascript', 'jscript', 'livescript', 'vbscript', 'data',
        'about', 'mocha'
    ])

    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self.fed = []
        self.ignore = False

    def handle_starttag(self, tag, attrs):
        if tag == 'script':
            self.ignore = True
            return
        elif self.ignore:
            return
        remove_css = self.DISALLOWED_STYLE_PATTERN.sub
        self.fed.extend(('<', tag))
        disallowed_schemes = tuple(scheme + ':'
                                   for scheme in self.DISALLOWED_SCHEMES)
        self.fed.extend(
            chunk
            for name, value in attrs
            if not name.startswith('on')
            for chunk in (
                [' ', name]
                if value is None else
                [
                    ' ', name, '="', cgi.escape(
                        ('' if value.startswith(disallowed_schemes) else value)
                        if name == 'href' else
                        (remove_css('\\1', value) if name == 'style' else value)
                    ), '"'
                ]
            )
        )
        self.fed.append('>')

    def handle_endtag(self, tag):
        if tag == 'script':
            self.ignore = False
            return
        self.fed.extend(('</', tag, '>'))

    def handle_data(self, d):
        if self.ignore:
            return
        self.fed.append(d)

    def handle_entityref(self, name):
        if self.ignore:
            return
        self.fed.extend(('&', name, ';'))

    def handle_charref(self, name):
        if self.ignore:
            return
        self.fed.extend(('&#' + name + ';'))

    def handle_comment(self, data):
        if self.ignore:
            return
        self.fed.extend(('<!-- ', data, ' -->'))
